from __future__ import annotations

import re
from typing import Any

from attr import dataclass
import numpy as np
import pandas as pd

from allotropy.allotrope.models.shared.definitions.custom import (
    TQuantityValueDalton,
    TQuantityValueMicrogramPerMilliliter,
    TQuantityValueMicroliterPerMinute,
    TQuantityValueNanomolar,
    TQuantityValueResponseUnit,
    TQuantityValueSecondTime,
)
from allotropy.allotrope.schema_mappers.adm.binding_affinity_analyzer.benchling._2024._12.binding_affinity_analyzer import (
    DeviceControlDocument,
)
from allotropy.exceptions import AllotropeConversionError
from allotropy.parsers.cytiva_biacore_insight import constants
from allotropy.parsers.cytiva_biacore_insight.cytiva_biacore_insight_reader import (
    CytivaBiacoreInsightReader,
)
from allotropy.parsers.utils.pandas import (
    df_to_series_data,
    drop_df_rows_while,
    map_rows,
    parse_header_row,
    SeriesData,
    split_dataframe,
)
from allotropy.parsers.utils.uuids import random_uuid_str
from allotropy.parsers.utils.values import (
    assert_not_none,
    quantity_or_none,
)
from allotropy.types import DictType


def _clean_custom_info(custom_info: dict[str, Any]) -> dict[str, Any]:
    """Remove None and NaN values from custom_info dictionaries."""

    def is_valid_value(value: Any) -> bool:
        if value is None:
            return False
        # Check for NaN in plain float values
        if isinstance(value, float) and np.isnan(value):
            return False
        # Check for NaN in quantity objects (objects with 'value' attribute or dicts with 'value' key)
        if hasattr(value, "value"):
            val = value.value
            if isinstance(val, float) and np.isnan(val):
                return False
        elif isinstance(value, dict) and "value" in value:
            val = value["value"]
            if isinstance(val, float) and np.isnan(val):
                return False
        return True

    return {key: value for key, value in custom_info.items() if is_valid_value(value)}


def _get_capture_custom_info(
    channel_data: pd.DataFrame, capture_1_solution: str | None
) -> dict[str, Any | None]:
    """
    Dynamically extract all Capture columns (Capture 1-5) that exist in the data.

    Args:
        channel_data: DataFrame with channel data
        capture_1_solution: Pre-extracted Capture 1 Solution value (for backwards compatibility)

    Returns:
        Dictionary with all available Capture fields
    """
    capture_info: dict[str, Any | None] = {}

    # OPTIMIZATION: Get column set once for fast lookups
    columns_set = set(channel_data.columns)

    # Check for Capture 1-5 (some files have multiple Capture sets)
    for capture_num in range(1, 6):
        solution_col = f"Capture {capture_num} Solution"

        # Special handling for Capture 1 Solution (use pre-extracted value if provided)
        if capture_num == 1 and capture_1_solution is not None:
            capture_info["Capture 1 Solution"] = capture_1_solution
        elif solution_col in columns_set:
            capture_info[solution_col] = _first_not_null_or_none(
                channel_data[solution_col]
            )

        # Add other Capture fields if they exist
        for field in ["Plate id", "Position", "Control type"]:
            col_name = f"Capture {capture_num} {field}"
            if col_name in columns_set:
                capture_info[col_name] = _first_not_null_or_none(channel_data[col_name])

    return capture_info


def _get_table_from_dataframe(df: pd.DataFrame, split_on: str) -> pd.DataFrame:
    data_table = drop_df_rows_while(df, lambda row: row[0] != split_on)
    return parse_header_row(
        assert_not_none(data_table).replace(np.nan, None)
    ).reset_index(drop=True)


def _first_not_null_or_none(series: pd.Series[Any]) -> str | None:
    """Get the first non-null value from a Series or None."""
    return str(series[idx]) if (idx := series.first_valid_index()) is not None else None


@dataclass(frozen=True)
class RunMetadata:
    sensor_chip_id: str
    sensor_chip_type: str
    lot_number: str
    model_number: str
    equipment_serial_number: str
    measurement_time: str
    number_of_cycles: float
    data_collection_rate: float
    running_buffer: str
    measurement_end_time: str
    method_name: str
    compartment_temperature: float | None
    run_table: pd.DataFrame

    @staticmethod
    def create_runs(runs_data: pd.DataFrame) -> list[RunMetadata]:
        runs_data = runs_data.reset_index(drop=True)
        run, rest = split_dataframe(
            runs_data, lambda row: str(row[0]).startswith("Run")
        )
        runs = [run]
        while rest is not None:
            run, rest = split_dataframe(rest, lambda row: str(row[0]).startswith("Run"))
            runs.append(run)
        if len(runs) > 1:
            msg = "Instrument file contains multiple runs. Only single runs are supported at the moment."
            raise AllotropeConversionError(msg)
        return [RunMetadata.create(run) for run in runs if not run.empty]

    @staticmethod
    def create(run_data: pd.DataFrame) -> RunMetadata:
        def get_run_information(idx: int) -> SeriesData:
            idx1, idx2 = idx + 1, idx + 4
            rd1 = run_data.loc[idx1:idx2][[3, 4]].rename(columns={3: "n", 4: "v"})
            rd2 = run_data.loc[idx1:idx2][[6, 7]].rename(columns={6: "n", 7: "v"})
            rd3 = run_data.loc[idx1:idx2][[10, 11]].rename(columns={10: "n", 11: "v"})
            return df_to_series_data(parse_header_row(pd.concat((rd1, rd2, rd3)).T))

        run_table_start = 0
        chip_data = SeriesData()
        run_information = SeriesData()
        run_table = pd.DataFrame()
        compartment_temperature = None
        for index, row in run_data.iterrows():
            idx = int(str(index))
            if row[0] == "Chip information":
                chip_data = df_to_series_data(
                    parse_header_row(run_data.loc[idx + 1 : idx + 4][[0, 1]].T)
                )
                run_information = get_run_information(idx)
            elif row[0] == "Flow cell":
                run_table_start = idx
            elif row[0] == "Method summary":
                run_table = (
                    parse_header_row(run_data.loc[run_table_start : idx - 1])
                    .dropna(how="all")
                    .reset_index(drop=True)
                )
                run_table = run_table.loc[:, ~run_table.columns.str.contains("^nan$")]
            elif row[0] == "Set temperatures 1":
                if match := re.match(constants.COMPARTMENT_TEMP_REGEX, row[1]):
                    compartment_temperature = float(match.group(1))

        return RunMetadata(
            sensor_chip_id=chip_data[str, "Chip id"],
            sensor_chip_type=chip_data[str, "Chip type"],
            lot_number=chip_data[str, "Lot number"],
            model_number=run_information[str, "Instrument type"],
            equipment_serial_number=run_information[str, "Instrument id"],
            measurement_time=run_information[str, "Start"],
            number_of_cycles=run_information[float, "Cycles"],
            data_collection_rate=run_information[float, "Data collection rate"],
            running_buffer=run_information[str, "Running buffer"],
            measurement_end_time=run_information[str, "End"],
            method_name=run_information[str, "Method"],
            compartment_temperature=compartment_temperature,
            run_table=run_table,
        )


@dataclass(frozen=True)
class BiacoreInsightMetadata:
    file_name: str
    unc_path: str
    software_name: str
    software_version: str
    analyst: str
    analytical_method_id: str
    runs: list[RunMetadata]
    data_processing_document: DictType | None = None

    @staticmethod
    def create(reader: CytivaBiacoreInsightReader) -> BiacoreInsightMetadata:

        properties = reader.data["Properties"]
        properties, runs_data = split_dataframe(
            properties, lambda row: str(row[0]).startswith("Run")
        )
        if runs_data is None or runs_data.empty:
            msg = "No run data found in the properties section of instrument file."
            raise AllotropeConversionError(msg)
        for index, row in properties.iterrows():
            idx = int(str(index))
            if row[0] == "Evaluation":
                evaluation_data = df_to_series_data(
                    parse_header_row(properties.loc[idx + 1 : idx + 4][[0, 1]].T)
                )
            elif row[0] == "Software":
                software_data = df_to_series_data(
                    parse_header_row(properties.loc[idx + 1 : idx + 2][[0, 1]].T)
                )

        data_processing_doc = BiacoreInsightMetadata.get_data_processing_doc(reader)

        return BiacoreInsightMetadata(
            file_name=evaluation_data[str, "Name"],
            unc_path=evaluation_data[str, "Path"],
            software_name=software_data[str, "Name"],
            software_version=software_data[str, "Version"],
            analyst=evaluation_data[str, "Modified by"],
            analytical_method_id=evaluation_data[str, "Name"],
            runs=RunMetadata.create_runs(runs_data),
            data_processing_document=data_processing_doc,
        )

    @staticmethod
    def get_data_processing_doc(reader: CytivaBiacoreInsightReader) -> DictType | None:
        table_names = [
            "QC - Capture baseline",
            "QC - Capture level",
            "QC - Binding to reference",
        ]
        qc_tables = [table for table in table_names if table in reader.data]
        qc_df = None
        while qc_df is None and qc_tables:
            qc_table = qc_tables.pop(0)
            _, qc_df = split_dataframe(
                reader.data[qc_table],
                lambda row: row[0] == "Blank subtraction",
                include_split_row=True,
            )

        if qc_df is None or qc_df.empty:
            return None

        qc_metadata = df_to_series_data(
            parse_header_row(qc_df.dropna(how="all", axis=0).T[0:2])
        )
        metadata_keys = [
            "Blank subtraction",
            "Molecular weight adjustment",
            "Capture/ligand adjustment",
            "Adjustment for controls",
            "Curve analysis",
        ]
        return {key: qc_metadata.get(str, key) for key in metadata_keys}


@dataclass(frozen=True)
class KineticsData:
    acceptance_state: str | None = None
    curve_markers: str | None = None
    kinetics_model: str | None = None
    binding_on_rate_measurement_datum: float | None = None
    binding_off_rate_measurement_datum: float | None = None
    equilibrium_dissociation_constant: float | None = None
    maximum_binding_capacity: float | None = None
    kinetics_chi_squared: float | None = None
    tc: float | None = None
    offset: float | None = None  # Affinity-specific field
    is_affinity_measurement: bool = False  # Track if this is from Affinity analysis

    @staticmethod
    def create(kinetics_data: SeriesData) -> KineticsData:
        # Detect if this is affinity data by checking which chi-squared column is present
        kinetics_chi = kinetics_data.get(float, "Kinetics Chi² (RU²)")
        affinity_chi = kinetics_data.get(float, "Affinity Chi² (RU²)")
        is_affinity = kinetics_chi is None and affinity_chi is not None

        return KineticsData(
            acceptance_state=kinetics_data.get(str, "Acceptance state"),
            curve_markers=kinetics_data.get(str, "Curve markers"),
            # Try "Kinetics model" first, fall back to "Affinity model"
            kinetics_model=(
                kinetics_data.get(str, "Kinetics model")
                or kinetics_data.get(str, "Affinity model")
            ),
            binding_on_rate_measurement_datum=kinetics_data.get(float, "ka (1/Ms)"),
            binding_off_rate_measurement_datum=kinetics_data.get(float, "kd (1/s)"),
            equilibrium_dissociation_constant=kinetics_data.get(float, "KD (M)"),
            maximum_binding_capacity=kinetics_data.get(float, "Rmax (RU)"),
            # Try "Kinetics Chi²" first, fall back to "Affinity Chi²"
            kinetics_chi_squared=kinetics_chi or affinity_chi,
            tc=kinetics_data.get(float, "tc"),
            offset=kinetics_data.get(float, "offset (RU)"),  # Affinity-specific
            is_affinity_measurement=is_affinity,
        )


class EvaluationKinetics:
    _data: dict[str, KineticsData]

    def __init__(self, kinetics_table: pd.DataFrame) -> None:
        def _get_key(row: pd.Series[Any]) -> str:
            # Handle both "Channel" and "Flow cell" columns
            channel_or_flowcell = row.get("Channel", row.get("Flow cell", ""))
            # For affinity data, we might have multiple capture solutions
            # Build key with all available capture solutions
            capture_solution = row.get("Capture 1 Solution", "")
            analyte_solution = row.get("Analyte 1 Solution", "")
            return f"{channel_or_flowcell} {capture_solution} {analyte_solution}"

        self._data = {}
        if not kinetics_table.empty and len(kinetics_table.columns) > 0:
            self._data = {
                _get_key(row): KineticsData.create(SeriesData(row))
                for _, row in kinetics_table.iterrows()
            }

    def get_data(
        self,
        channel: int | str,
        capture_solution: str | None,
        analyte_solution: str | None,
    ) -> KineticsData:
        empty = KineticsData()
        if capture_solution is None or analyte_solution is None:
            return empty
        return self._data.get(f"{channel} {capture_solution} {analyte_solution}", empty)


@dataclass(frozen=True)
class ReportPointData:
    identifier: str
    identifier_role: str
    absolute_resonance: float
    time_setting: float
    relative_resonance: float | None
    step_name: str | None
    step_purpose: str | None
    window: float | None
    baseline: str | None
    lrsd: float | None
    slope: float | None
    standard_deviation: float | None

    @staticmethod
    def create(row: SeriesData) -> ReportPointData:
        return ReportPointData(
            identifier=random_uuid_str(),
            identifier_role=row[str, "Name"],
            absolute_resonance=row[float, "Absolute response (RU)"],
            time_setting=row[float, "Time (s)"],
            relative_resonance=row.get(float, "Relative response (RU)"),
            step_name=row.get(str, "Name"),
            step_purpose=row.get(str, "Step purpose"),
            window=row.get(float, "Window (s)"),
            baseline=row.get(str, "Baseline"),
            lrsd=row.get(float, "LRSD"),
            slope=row.get(float, "Slope (RU/s)"),
            standard_deviation=row.get(float, "Standard deviation"),
        )


@dataclass(frozen=True)
class MeasurementData:
    identifier: str
    sample_identifier: str
    method_name: str
    ligand_identifier: str | None
    device_control_document: list[DeviceControlDocument]
    sample_custom_info: dict[str, Any | None]
    kinetics: KineticsData
    report_point_data: list[ReportPointData]

    @staticmethod
    def create(
        channel_data: pd.DataFrame,
        metadata: BiacoreInsightMetadata,
        evaluation_kinetics: EvaluationKinetics,
        grouping_column: str = "Channel",
    ) -> MeasurementData:
        identifier = random_uuid_str()
        run_metadata = metadata.runs[0]
        first_row_data = SeriesData(channel_data.iloc[0])
        run = first_row_data[int, "Run"]
        cycle_number = first_row_data[int, "Cycle"]

        # Handle both "Channel" and "Flow cell" as grouping columns
        # When using Flow cell, convert to string since it can be "2-1" format
        channel: int | str
        if grouping_column == "Channel":
            channel = first_row_data[int, "Channel"]
        else:
            # Flow cell can be string like "2-1", store as string
            channel = str(first_row_data.get(str, grouping_column, ""))

        run_table = run_metadata.run_table
        device_control_document = []
        for flow_cell, flow_cell_data in channel_data.groupby("Flow cell"):
            # Filter run_table using the appropriate column
            if grouping_column == "Channel" and "Channel" in run_table.columns:
                run_data = run_table[
                    (run_table["Flow cell"].astype(str) == str(flow_cell))
                    & (run_table["Channel"] == channel)
                ]
            else:
                # For new format, just filter by flow cell
                run_data = run_table[
                    run_table["Flow cell"].astype(str) == str(flow_cell)
                ]
            run_info = SeriesData() if run_data.empty else df_to_series_data(run_data)
            # All the info needed in the device control document is the same for all flow cell data
            flow_cell_info = SeriesData(flow_cell_data.iloc[0])
            device_control_document.append(
                DeviceControlDocument(
                    device_type=constants.DEVICE_TYPE,
                    flow_cell_identifier=str(flow_cell),
                    sample_temperature_setting=flow_cell_info.get(
                        float, "Temperature (°C)"
                    ),
                    device_control_custom_info=_clean_custom_info(
                        {
                            "Analyte 1 Contact time": quantity_or_none(
                                TQuantityValueSecondTime,
                                flow_cell_info.get(float, "Analyte 1 Contact time (s)"),
                            ),
                            "Analyte 1 Dissociation time": quantity_or_none(
                                TQuantityValueSecondTime,
                                flow_cell_info.get(
                                    float, "Analyte 1 Dissociation time (s)"
                                ),
                            ),
                            "Analyte 1 Flow rate": quantity_or_none(
                                TQuantityValueMicroliterPerMinute,
                                flow_cell_info.get(
                                    float, "Analyte 1 Flow rate (µl/min)"
                                ),
                            ),
                            "Regeneration 1 Contact time": quantity_or_none(
                                TQuantityValueSecondTime,
                                flow_cell_info.get(
                                    float, "Regeneration 1 Contact time (s)"
                                ),
                            ),
                            "Regeneration 1 Flow rate": quantity_or_none(
                                TQuantityValueMicroliterPerMinute,
                                flow_cell_info.get(
                                    float, "Regeneration 1 Flow rate (µl/min)"
                                ),
                            ),
                            "Included": run_info.get(str, "Included"),
                            "Sensorgram type": flow_cell_info.get(
                                str,
                                "Sensorgram type",
                                run_info.get(str, "Sensorgram type"),
                            ),
                            "Level": quantity_or_none(
                                TQuantityValueResponseUnit,
                                run_info.get(float, "Level (RU)"),
                            ),
                        }
                    ),
                )
            )

        # Capture columns may not exist in all file formats
        capture_solution = None
        if "Capture 1 Solution" in channel_data.columns:
            capture_solution = _first_not_null_or_none(
                channel_data["Capture 1 Solution"]
            )
        analyte_solution = first_row_data.get(str, "Analyte 1 Solution")

        return MeasurementData(
            identifier=identifier,
            sample_identifier=f"Run{run}_Cycle{cycle_number}",
            method_name=run_metadata.method_name,
            ligand_identifier=run_info.get(str, "Ligand"),
            device_control_document=device_control_document,
            sample_custom_info=_clean_custom_info(
                {
                    "Run": run,
                    "Cycle": cycle_number,
                    "Channel": channel,
                    "Analyte 1 Solution": analyte_solution,
                    "Analyte 1 Plate id": first_row_data.get(str, "Analyte 1 Plate id"),
                    "Analyte 1 Position": first_row_data.get(str, "Analyte 1 Position"),
                    "Analyte 1 Control type": first_row_data.get(
                        str, "Analyte 1 Control type"
                    ),
                    "Regeneration 1 Solution": first_row_data.get(
                        str, "Regeneration 1 Solution"
                    ),
                    "Regeneration 1 Plate id": first_row_data.get(
                        str, "Regeneration 1 Plate id"
                    ),
                    "Regeneration 1 Position": first_row_data.get(
                        str, "Regeneration 1 Position"
                    ),
                    "Regeneration 1 Control type": first_row_data.get(
                        str, "Regeneration 1 Control type"
                    ),
                    # Capture data may not be present in all file formats
                    # Dynamically add all Capture columns that exist (Capture 1-5)
                    **_get_capture_custom_info(channel_data, capture_solution),
                    "Analyte 1 Concentration": (
                        # Try nM first, then µg/ml
                        quantity_or_none(
                            TQuantityValueNanomolar,
                            first_row_data.get(float, "Analyte 1 Concentration (nM)"),
                        )
                        or quantity_or_none(
                            TQuantityValueMicrogramPerMilliliter,
                            first_row_data.get(
                                float, "Analyte 1 Concentration (µg/ml)"
                            ),
                        )
                    ),
                    "Analyte 1 Molecular weight": quantity_or_none(
                        TQuantityValueDalton,
                        first_row_data.get(float, "Analyte 1 Molecular weight (Da)"),
                    ),
                }
            ),
            kinetics=evaluation_kinetics.get_data(
                channel, capture_solution, analyte_solution
            ),
            report_point_data=map_rows(channel_data, ReportPointData.create),
        )


@dataclass(frozen=True)
class Data:
    metadata: BiacoreInsightMetadata
    cycles: dict[int, list[MeasurementData]]

    @staticmethod
    def create(reader: CytivaBiacoreInsightReader) -> Data:
        metadata = BiacoreInsightMetadata.create(reader)

        # OPTIMIZATION: Parse report point table once, not per cycle
        report_point_table = _get_table_from_dataframe(
            reader.data["Report point table"], split_on="Run"
        )

        # OPTIMIZATION: Determine grouping column once
        grouping_column = (
            "Channel" if "Channel" in report_point_table.columns else "Flow cell"
        )

        # OPTIMIZATION: Parse evaluation/affinity data once, not per cycle
        evaluation_kinetics = None
        if "Evaluation - Kinetics" in reader.data:
            evaluation_kinetics_table = _get_table_from_dataframe(
                reader.data["Evaluation - Kinetics"], split_on="Group"
            )
            evaluation_kinetics = EvaluationKinetics(evaluation_kinetics_table)
        else:
            # Look for Affinity sheets (e.g., "Affinity 1", "Affinity 2", etc.)
            affinity_sheets = [
                sheet for sheet in reader.data.keys() if sheet.startswith("Affinity")
            ]
            if affinity_sheets:
                # Use the first Affinity sheet found
                affinity_table = _get_table_from_dataframe(
                    reader.data[affinity_sheets[0]], split_on="Group"
                )
                evaluation_kinetics = EvaluationKinetics(affinity_table)

        # If no kinetics or affinity data found, create empty
        if evaluation_kinetics is None:
            evaluation_kinetics = EvaluationKinetics(pd.DataFrame())

        # OPTIMIZATION: Group all data by cycle once using pandas groupby
        # This is much faster than filtering per cycle
        cycles_dict = {}
        for cycle_number, cycle_group in report_point_table.groupby("Cycle"):
            # Convert cycle_number to int safely
            # pandas groupby can return various numeric types
            if isinstance(cycle_number, int | np.integer):
                cycle_int = int(cycle_number)
            else:
                cycle_int = int(float(str(cycle_number)))
            cycles_dict[cycle_int] = [
                MeasurementData.create(
                    channel_data, metadata, evaluation_kinetics, grouping_column
                )
                for _, channel_data in cycle_group.groupby(grouping_column)
            ]

        return Data(
            metadata=metadata,
            cycles=cycles_dict,
        )

    @staticmethod
    def create_measurements_for_cycle(
        cycle_number: int,
        metadata: BiacoreInsightMetadata,
        reader: CytivaBiacoreInsightReader,
    ) -> list[MeasurementData]:
        """
        DEPRECATED: This method is no longer used for performance reasons.
        All cycle processing is now done in Data.create() to avoid redundant parsing.
        Keeping this method for backward compatibility in case it's called externally.
        """
        # Report point table is the entry point
        report_point_table = _get_table_from_dataframe(
            reader.data["Report point table"], split_on="Run"
        )
        cycle_data = report_point_table[report_point_table["Cycle"] == cycle_number]

        # Handle optional kinetics/affinity data
        # Try "Evaluation - Kinetics" first, then look for "Affinity" sheets
        evaluation_kinetics = None
        if "Evaluation - Kinetics" in reader.data:
            evaluation_kinetics_table = _get_table_from_dataframe(
                reader.data["Evaluation - Kinetics"], split_on="Group"
            )
            evaluation_kinetics = EvaluationKinetics(evaluation_kinetics_table)
        else:
            # Look for Affinity sheets (e.g., "Affinity 1", "Affinity 2", etc.)
            affinity_sheets = [
                sheet for sheet in reader.data.keys() if sheet.startswith("Affinity")
            ]
            if affinity_sheets:
                # Use the first Affinity sheet found
                affinity_table = _get_table_from_dataframe(
                    reader.data[affinity_sheets[0]], split_on="Group"
                )
                evaluation_kinetics = EvaluationKinetics(affinity_table)

        # If no kinetics or affinity data found, create empty
        if evaluation_kinetics is None:
            evaluation_kinetics = EvaluationKinetics(pd.DataFrame())

        # Determine grouping column: use "Channel" if available, otherwise "Flow cell"
        grouping_column = "Channel" if "Channel" in cycle_data.columns else "Flow cell"

        return [
            MeasurementData.create(
                channel_data, metadata, evaluation_kinetics, grouping_column
            )
            for _, channel_data in cycle_data.groupby(grouping_column)
        ]



import re
from typing import ClassVar

import pandas as pd

from allotropy.named_file_contents import NamedFileContents
from allotropy.parsers.beckman_coulter_biomek.constants import (
    FileFormat,
    PIPETTING_COLUMNS,
    UNIFIED_PIPETTING_COLUMNS,
    UNIFIED_TRANSFER_COLUMNS,
)
from allotropy.parsers.lines_reader import CsvReader
from allotropy.parsers.utils.pandas import SeriesData
from allotropy.parsers.utils.values import assert_not_none


class BeckmanCoulterBiomekReader:
    SUPPORTED_EXTENSIONS = "csv,log"
    header: SeriesData
    data: pd.DataFrame
    file_format: FileFormat

    # Fixed column names based on headerless files - keeping for backwards compatibility
    FIXED_COLUMN_NAMES: ClassVar[list[str]] = UNIFIED_PIPETTING_COLUMNS

    def _looks_like_data_line(self, line: str) -> bool:
        """
        Check if a line looks like data rather than metadata.
        Data lines should have a date/time pattern as the first field.
        """
        parts = line.split(",")
        if len(parts) < 3:
            return False

        first_part = parts[0].strip()
        # Check if first part looks like a timestamp (contains numbers and slashes or dashes)
        # Examples: "08/15/2025 10:59:55", "9/4/24 14:15"
        timestamp_pattern = r"^\d{1,2}[/-]\d{1,2}[/-]\d{2,4}\s+\d{1,2}:\d{2}"
        return bool(re.match(timestamp_pattern, first_part))

    def _detect_file_format(self, data_lines: list[str], filename: str) -> FileFormat:
        """
        Detect the file format based on column count and filename patterns.
        """
        if not data_lines:
            # Default to unified pipetting for empty files
            return FileFormat.UNIFIED_PIPETTING

        # Count columns in first data line
        column_count = len(data_lines[0].split(","))

        # Check filename patterns first
        filename_lower = filename.lower()
        if "unifiedtransfer" in filename_lower:
            return FileFormat.UNIFIED_TRANSFER
        elif "pipetting" in filename_lower and "unified" not in filename_lower:
            return FileFormat.PIPETTING

        # Fall back to column count detection
        if column_count == 13:
            return FileFormat.UNIFIED_TRANSFER
        elif column_count == 9:
            return FileFormat.PIPETTING
        else:
            # Default to unified pipetting (11 columns)
            return FileFormat.UNIFIED_PIPETTING

    def _get_column_names(self, file_format: FileFormat) -> list[str]:
        """Get the appropriate column names for the detected file format."""
        if file_format == FileFormat.UNIFIED_TRANSFER:
            return UNIFIED_TRANSFER_COLUMNS
        elif file_format == FileFormat.PIPETTING:
            return PIPETTING_COLUMNS
        else:
            return UNIFIED_PIPETTING_COLUMNS

    def __init__(self, named_file_contents: NamedFileContents) -> None:
        reader = CsvReader.create(named_file_contents)

        # Check if this file has a header line by looking for "Well Index"
        has_header_line = any("Well Index" in line for line in reader.lines)

        if has_header_line:
            header_lines = list(reader.pop_until(".*Well Index.*"))
            header_dict = {}
            for line in header_lines:
                split = line.split(",")[0].split("=", maxsplit=1)
                if len(split) != 2:
                    continue
                header_dict[split[0].strip()] = split[1].strip()

            self.header = SeriesData(pd.Series(header_dict))
            self.data = assert_not_none(
                reader.pop_csv_block_as_df(header="infer"), "Cannot parse empty dataset"
            )
            # For files with headers, detect format from data
            data_lines = [",".join(str(val) for val in row) for row in self.data.values]
            self.file_format = self._detect_file_format(
                data_lines, named_file_contents.original_file_path
            )
        else:
            # Handle headerless files
            header_lines = []
            data_lines = []

            for raw_line in reader.lines:
                line = raw_line.strip()
                if not line:
                    continue

                if self._looks_like_data_line(line):
                    data_lines.append(line)
                else:
                    header_lines.append(line)

            # Detect file format before creating DataFrame
            self.file_format = self._detect_file_format(
                data_lines, named_file_contents.original_file_path
            )
            column_names = self._get_column_names(self.file_format)

            # Parse header metadata
            header_dict = {}
            for line in header_lines:
                split = line.split(",")[0].split("=", maxsplit=1)
                if len(split) != 2:
                    continue
                header_dict[split[0].strip()] = split[1].strip()

            self.header = SeriesData(pd.Series(header_dict))

            # Create a new reader with just the data lines
            data_reader = CsvReader(data_lines)
            self.data = assert_not_none(
                data_reader.pop_csv_block_as_df(header=None, names=column_names),
                "Cannot parse empty dataset",
            )



from __future__ import annotations

from collections.abc import Iterator
from dataclasses import dataclass
from itertools import chain

from allotropy.calcdocs.extractor import Element
from allotropy.calcdocs.view import Keys, ViewData
from allotropy.parsers.utils.calculated_data_documents.definition import (
    CalculatedDocument,
    DataSource,
)
from allotropy.parsers.utils.uuids import random_uuid_str


@dataclass(frozen=True)
class CalcDocsConfig:
    configs: list[CalculatedDataConfig]

    def construct(self) -> list[CalculatedDocument]:
        cache: dict[str, CalculatedDocument | None] = {}
        return list(chain(*[config.construct(cache) for config in self.configs]))


@dataclass(frozen=True)
class CalculatedDataConfig:
    name: str
    value: str
    view_data: ViewData
    source_configs: tuple[CalculatedDataConfig | MeasurementConfig, ...]
    unit: str | None = None
    description: str | None = None
    # Optional key to pull a dynamic description from the element data
    description_value_key: str | None = None
    required: bool = False

    def iter_data_sources(
        self,
        parent_keys: Keys,
        _: list[Element],
        cache: dict[str, CalculatedDocument | None],
    ) -> Iterator[DataSource]:
        keys = self.view_data.filter_keys(parent_keys)
        item = self.view_data.get_item(keys)
        sub_keys_iterator = item.iter_keys() if isinstance(item, ViewData) else [Keys()]

        for sub_keys in sub_keys_iterator:
            if calc_doc := self.get_calc_doc(keys.append(sub_keys), cache):
                yield DataSource(
                    feature=calc_doc.name,
                    reference=calc_doc,
                    value=calc_doc.value,
                )

    def _get_calc_doc_inner(
        self,
        keys: Keys,
        cache: dict[str, CalculatedDocument | None],
    ) -> CalculatedDocument | None:
        elements = self.view_data.get_leaf_items(keys)
        value = elements[0].get_float_or_none(self.value)

        if value is None:
            return None

        data_sources = []
        for s_config in self.source_configs:
            if sources := list(s_config.iter_data_sources(keys, elements, cache)):
                data_sources.extend(sources)
            elif s_config.required:
                return None

        if not data_sources:
            return None

        if self.name == "B22 goodness of fit":
            pass
        return CalculatedDocument(
            uuid=random_uuid_str(),
            name=self.name,
            value=value,
            data_sources=data_sources,
            unit=self.unit,
            description=(
                self.description
                if self.description is not None
                else (
                    elements[0].get_str_or_none(self.description_value_key)
                    if self.description_value_key
                    else None
                )
            ),
        )

    def get_cache_key(self, keys: Keys) -> str:
        return f"{self.name} {self.value} {keys}"

    def get_calc_doc(
        self,
        keys: Keys,
        cache: dict[str, CalculatedDocument | None],
    ) -> CalculatedDocument | None:
        key = self.get_cache_key(keys)
        if result := cache.get(key):
            return result
        result = self._get_calc_doc_inner(keys, cache)
        cache[key] = result
        return result

    def construct(
        self, cache: dict[str, CalculatedDocument | None]
    ) -> list[CalculatedDocument]:
        return [
            calc_doc
            for keys in self.view_data.iter_keys()
            if (calc_doc := self.get_calc_doc(keys, cache))
        ]


@dataclass(frozen=True)
class MeasurementConfig:
    name: str
    value: str
    required: bool = False

    def iter_data_sources(
        self,
        _: Keys,
        elements: list[Element],
        __: dict[str, CalculatedDocument | None],
    ) -> Iterator[DataSource]:
        for element in elements:
            if (value := element.get_float_or_none(self.value)) is not None:
                yield DataSource(
                    feature=self.name,
                    reference=element,
                    value=value,
                )



from allotropy.allotrope.models.adm.plate_reader.rec._2025._03.plate_reader import (
    Model,
)
from allotropy.allotrope.schema_mappers.adm.plate_reader.rec._2025._03.plate_reader import (
    Data,
    Mapper,
)
from allotropy.named_file_contents import NamedFileContents
from allotropy.parsers.agilent_gen5.agilent_gen5_reader import AgilentGen5Reader
from allotropy.parsers.agilent_gen5.agilent_gen5_structure import (
    create_metadata,
    process_all_reads,
)
from allotropy.parsers.release_state import ReleaseState
from allotropy.parsers.vendor_parser import VendorParser


class AgilentGen5Parser(VendorParser[Data, Model]):
    DISPLAY_NAME = "Agilent Gen5"
    RELEASE_STATE = ReleaseState.RECOMMENDED
    SUPPORTED_EXTENSIONS = AgilentGen5Reader.SUPPORTED_EXTENSIONS
    SCHEMA_MAPPER = Mapper

    def create_data(self, named_file_contents: NamedFileContents) -> Data:
        reader = AgilentGen5Reader(named_file_contents)
        context = reader.extract_data_context(named_file_contents.original_file_path)

        # Process all reads and merge results
        measurement_groups, calculated_data = process_all_reads(reader, context)

        return Data(
            metadata=create_metadata(context.header_data),
            measurement_groups=measurement_groups,
            calculated_data=calculated_data,
        )



from allotropy.calcdocs.extractor import Element, Extractor
from allotropy.parsers.msd_workbench.calculdated_data_structure import (
    CalculatedDataMeasurementStructure,
)


class MsdWorkbenchExtractor(Extractor[CalculatedDataMeasurementStructure]):
    @classmethod
    def to_element(
        cls, calc_data_structure: CalculatedDataMeasurementStructure
    ) -> Element:
        return Element(
            uuid=calc_data_structure.measurement.identifier,
            data={
                "type_": calc_data_structure.measurement.type_.value,
                "uuid": calc_data_structure.measurement.identifier,
                "luminescence": calc_data_structure.measurement.luminescence,
                "sample_identifier": calc_data_structure.measurement.sample_identifier,
                "location_identifier": calc_data_structure.measurement.location_identifier,
                "well_location_identifier": calc_data_structure.measurement.well_location_identifier,
                "well_plate_identifier": calc_data_structure.measurement.well_plate_identifier,
                "device_type": calc_data_structure.measurement.device_type,
                "detection_type": calc_data_structure.measurement.detection_type,
                "mass_concentration": calc_data_structure.measurement.mass_concentration,
                "sample_role_type": calc_data_structure.measurement.sample_role_type.value
                if calc_data_structure.measurement.sample_role_type
                else None,
                "assay_identifier": calc_data_structure.measurement.measurement_custom_info.get(
                    "assay identifier"
                )
                if calc_data_structure.measurement.measurement_custom_info
                else None,
                "adjusted_signal": calc_data_structure.adjusted_signal,
                "mean": calc_data_structure.mean,
                "adj_sig_mean": calc_data_structure.adj_sig_mean,
                "fit_statistic_rsquared": calc_data_structure.fit_statistic_rsquared,
                "cv": calc_data_structure.cv,
                "percent_recovery": calc_data_structure.percent_recovery,
                "percent_recovery_mean": calc_data_structure.percent_recovery_mean,
                "calc_concentration": calc_data_structure.calc_concentration,
                "calc_conc_mean": calc_data_structure.calc_conc_mean,
                "calc_conc_cv": calc_data_structure.calc_conc_cv,
            },
        )



from __future__ import annotations

from collections import defaultdict
from dataclasses import dataclass
from pathlib import Path
from typing import Any

import pandas as pd

from allotropy.allotrope.schema_mappers.adm.plate_reader.rec._2024._06.plate_reader import (
    Measurement,
    MeasurementGroup,
    MeasurementType,
    Metadata,
)
from allotropy.exceptions import AllotropyParserError
from allotropy.parsers.constants import (
    DEFAULT_EPOCH_TIMESTAMP,
    get_well_count_by_well_ids,
    NOT_APPLICABLE,
)
from allotropy.parsers.msd_workbench.constants import (
    DETECTION_TYPE,
    DEVICE_TYPE,
    SAMPLE_ROLE_TYPE_MAPPING,
    SOFTWARE_NAME,
)
from allotropy.parsers.msd_workbench.msd_workbench_calculated_data_mapping import (
    CalculatedDataColumns,
)
from allotropy.parsers.utils.pandas import map_rows, SeriesData
from allotropy.parsers.utils.uuids import random_uuid_str


@dataclass(frozen=True)
class PlateData:
    measurement_time: str
    plate_well_count: int
    well_plate_id: str
    well_data: pd.DataFrame

    @staticmethod
    def create(
        data: pd.DataFrame,
        well_plate_id: str,
    ) -> PlateData:
        plate_well_count = PlateData._get_plate_well_count(data)
        if not plate_well_count:
            msg = "Could not determine plate well count"
            raise AllotropyParserError(msg)

        return PlateData(
            measurement_time=DEFAULT_EPOCH_TIMESTAMP,
            well_plate_id=well_plate_id,
            plate_well_count=plate_well_count,
            well_data=data,
        )

    @staticmethod
    def _get_plate_well_count(data: pd.DataFrame) -> int | None:
        return get_well_count_by_well_ids(
            well_identifiers=[int(well[-1]) for well in data["Well"]],
            well_locations=data["Well"].astype(str).tolist(),
        )


def create_metadata(file_name: str) -> Metadata:
    asm_file_identifier = Path(file_name).with_suffix(".json")
    return Metadata(
        file_name=file_name.rsplit("\\", 1)[-1],
        unc_path=file_name,
        software_name=SOFTWARE_NAME,
        device_identifier=NOT_APPLICABLE,
        model_number=NOT_APPLICABLE,
        asm_file_identifier=asm_file_identifier.name,
        data_system_instance_id=NOT_APPLICABLE,
    )


def create_measurement_groups(plate_data: PlateData) -> list[MeasurementGroup]:
    def map_measurement(row: SeriesData) -> Measurement:
        sample_id = f"{row[str, 'Sample']}_{row[str, 'Well']}"
        custom_info = {
            "detection range": row.get(str, "Detection Range"),
            "assay identifier": row.get(str, "Assay"),
        }
        # Try both "Dilution Factor" and "Dilution" for backwards compatibility
        dilution_value = row.get(int, "Dilution Factor")
        if dilution_value is None:
            dilution_value = row.get(int, "Dilution")

        return Measurement(
            type_=MeasurementType.LUMINESCENCE,
            identifier=random_uuid_str(),
            luminescence=row[float, "Signal"],
            sample_identifier=sample_id,
            location_identifier=row[str, "Spot"],
            well_location_identifier=row[str, "Well"],
            well_plate_identifier=plate_data.well_plate_id,
            device_type=DEVICE_TYPE,
            detection_type=DETECTION_TYPE,
            mass_concentration=row.get(float, "Concentration"),
            sample_role_type=SAMPLE_ROLE_TYPE_MAPPING.get(
                row[str, "Sample"][0].lower()
            ),
            sample_custom_info={
                "dilution factor setting": dilution_value,
            },
            measurement_custom_info={
                **custom_info,
                **_filter_calculated_data_fields(row.get_unread()),
            },
        )

    measurements = map_rows(plate_data.well_data, map_measurement)

    grouped_measurements = defaultdict(list)
    for measurement in measurements:
        grouped_measurements[measurement.sample_identifier].append(measurement)

    return [
        MeasurementGroup(
            measurement_time=plate_data.measurement_time,
            plate_well_count=plate_data.plate_well_count,
            measurements=group,
        )
        for group in grouped_measurements.values()
    ]


def _filter_calculated_data_fields(unread_data: dict[str, Any]) -> dict[str, Any]:
    return {
        key: value
        for key, value in unread_data.items()
        if key not in {col.value for col in CalculatedDataColumns}
    }



from __future__ import annotations

from abc import ABC, abstractmethod
from collections import defaultdict
from dataclasses import dataclass
from io import StringIO
import math
from pathlib import Path
import re
from typing import Any, cast

import pandas as pd

from allotropy.allotrope.models.shared.definitions.definitions import (
    FieldComponentDatatype,
)
from allotropy.allotrope.models.shared.definitions.units import (
    MilliAbsorbanceUnit,
    Nanometer,
    UNITLESS,
)
from allotropy.allotrope.schema_mappers.adm.plate_reader.rec._2025._03.plate_reader import (
    ErrorDocument,
    Measurement,
    MeasurementGroup,
    MeasurementType,
    Metadata,
    ScanPositionSettingPlateReader,
)
from allotropy.allotrope.schema_mappers.data_cube import DataCube, DataCubeComponent
from allotropy.exceptions import (
    AllotropeConversionError,
    AllotropyParserError,
)
from allotropy.parsers.agilent_gen5.constants import (
    ALPHALISA_FLUORESCENCE_FOUND,
    DATA_SOURCE_FEATURE_VALUES,
    DEFAULT_SOFTWARE_NAME,
    DEVICE_TYPE,
    ELAPSED_TIME,
    EMISSION_KEY,
    EMISSION_START_KEY,
    EXCITATION_KEY,
    EXCITATION_START_KEY,
    FILENAME_REGEX,
    FIXED_EMISSION_KEY,
    FIXED_EXCITATION_KEY,
    GAIN_KEY,
    LIGHT_DIRECTIONS,
    MEASUREMENTS_DATA_POINT_KEY,
    MIRROR_KEY,
    NAN_EMISSION_EXCITATION,
    NO_MEASUREMENTS_ERROR,
    OPTICS_KEY,
    PATHLENGTH_CORRECTION_KEY,
    READ_DATA_MEASUREMENT_ERROR,
    READ_HEIGHT_KEY,
    READ_MODE_UNITS,
    READ_SPEED_KEY,
    ReadMode,
    ReadType,
    SECONDS,
    UNSUPPORTED_READ_MODE_ERROR,
    UNSUPPORTED_READ_TYPE_ERROR,
    WAVELENGTHS_KEY,
)
from allotropy.parsers.constants import (
    get_well_count_by_well_ids,
    NEGATIVE_ZERO,
    NOT_APPLICABLE,
    POSSIBLE_WELL_COUNTS,
    round_to_nearest_well_count,
)
from allotropy.parsers.lines_reader import SectionLinesReader
from allotropy.parsers.utils.calculated_data_documents.definition import (
    CalculatedDocument,
    DataSource,
    Referenceable,
)
from allotropy.parsers.utils.pandas import read_csv, SeriesData
from allotropy.parsers.utils.uuids import random_uuid_str
from allotropy.parsers.utils.values import (
    try_float,
    try_float_or_none,
    try_non_nan_float_or_none,
)


@dataclass(frozen=True)
class HeaderData:
    datetime: str
    software_version: str
    experiment_file_path: str | None
    protocol_file_path: str | None
    well_plate_identifier: str | None
    model_number: str | None
    equipment_serial_number: str | None
    additional_data: dict[str, str | float | None]
    plate_well_count: float
    file_name: str
    unc_path: str

    @classmethod
    def create(cls, data: SeriesData, file_path: str) -> HeaderData:
        file_name = Path(file_path).name
        matches = re.match(FILENAME_REGEX, file_name)
        plate_identifier = matches.groupdict()["plate_identifier"] if matches else None
        plate_well_count = cls._extract_well_count(data[str, "Plate Type"])
        return HeaderData(
            software_version=data[str, "Software Version"],
            experiment_file_path=data.get(str, "Experiment File Path:"),
            file_name=file_name,
            unc_path=file_path,
            protocol_file_path=data.get(str, "Protocol File Path:"),
            datetime=f'{data[str, "Date"]} {data[str, "Time"]}',
            well_plate_identifier=plate_identifier or data.get(str, "Plate Number"),
            model_number=data.get(str, "Reader Type:"),
            equipment_serial_number=data.get(str, "Reader Serial Number:"),
            plate_well_count=plate_well_count,
            additional_data=data.get_unread(),
        )

    @staticmethod
    def _extract_well_count(plate_type: str) -> float:
        match = re.search(
            rf"({'|'.join(str(count) for count in POSSIBLE_WELL_COUNTS)})", plate_type
        )
        if match:
            return float(match.group())
        return 0


@dataclass(frozen=True)
class FilterSet:
    gain: str
    emission: str | None = None
    excitation: str | None = None
    mirror: str | None = None
    optics: str | None = None
    light_direction: str | None = None

    @property
    def detector_wavelength_setting(self) -> float | None:
        if not self.emission:
            return None
        return try_non_nan_float_or_none(self.emission.split("/")[0])

    @property
    def detector_bandwidth_setting(self) -> float | None:
        if not self.emission:
            return None
        try:
            return try_non_nan_float_or_none(self.emission.split("/")[1])
        except IndexError:
            return None

    @property
    def excitation_wavelength_setting(self) -> float | None:
        if self.excitation:
            return try_non_nan_float_or_none(self.excitation.split("/")[0])
        return None

    @property
    def excitation_bandwidth_setting(self) -> float | None:
        if not self.excitation:
            return None
        try:
            return try_float(self.excitation.split("/")[1], "Excitation bandwith")
        except IndexError:
            return None

    @property
    def wavelength_filter_cutoff_setting(self) -> float | None:
        if self.mirror:
            return try_float(self.mirror.split(" ")[1], "Wavelength filter cutoff")
        return None

    @property
    def scan_position_setting(self) -> ScanPositionSettingPlateReader | None:
        position_lookup = {
            "Top": ScanPositionSettingPlateReader.top_scan_position__plate_reader_,
            "Bottom": ScanPositionSettingPlateReader.bottom_scan_position__plate_reader_,
        }
        position = self.optics
        if self.mirror:
            position = self.mirror.split(" ")[0]

        # TODO: Raise if position is not a valid lookup value
        return position_lookup.get(position) if position is not None else None


@dataclass
class DeviceControlData:
    step_label: str | None
    _list_data: dict[str, list[str]]
    _single_data: dict[str, str]
    is_polarization: bool = False

    LIST_KEYS = frozenset(
        {
            EMISSION_KEY,
            EXCITATION_KEY,
            FIXED_EXCITATION_KEY,
            FIXED_EMISSION_KEY,
            EMISSION_START_KEY,
            EXCITATION_START_KEY,
            OPTICS_KEY,
            GAIN_KEY,
            MIRROR_KEY,
            WAVELENGTHS_KEY,
        }
    )

    def __init__(self) -> None:
        self.step_label = None
        self._list_data = {key: [] for key in self.LIST_KEYS}
        self._single_data = {}

    def add(self, key: str, value: str | list[str]) -> None:
        if key in self.LIST_KEYS:
            if isinstance(value, str):
                self._list_data[key].append(value)
            else:
                self._list_data[key].extend(value)
        else:
            if not isinstance(value, str):
                msg = f"Unexpected list passed to key: {key}"
                raise AllotropyParserError(msg)
            self._single_data[key] = value

    def get_list(self, key: str) -> list[str]:
        return self._list_data.get(key, [])

    def get(self, key: str) -> str | None:
        return self._single_data.get(key)

    @classmethod
    def create(cls, lines: list[str], read_mode: ReadMode) -> DeviceControlData:

        device_control_data = DeviceControlData()

        for line in lines:
            strp_line = str(line.strip())
            if strp_line == "Fluorescence Polarization":
                device_control_data.is_polarization = True
                continue

            if strp_line.startswith("Read\t"):
                device_control_data.step_label = cls._get_step_label(line, read_mode)
                continue

            elif strp_line.startswith(WAVELENGTHS_KEY):
                wavelengths = strp_line.split(":  ")
                device_control_data.add(WAVELENGTHS_KEY, wavelengths[1].split(", "))
                continue

            # Handle excitation/emission wavelengths in special patterns
            elif "Fixed Excitation:" in strp_line:
                fixed_excitation = strp_line.split("Fixed Excitation:")[1].strip()
                device_control_data.add(FIXED_EXCITATION_KEY, fixed_excitation)
                continue

            elif "Fixed Emission:" in strp_line:
                fixed_emission = strp_line.split("Fixed Emission:")[1].strip()
                device_control_data.add(FIXED_EMISSION_KEY, fixed_emission)
                continue

            elif "Emission Start:" in strp_line:
                emission_start = strp_line.split("Emission Start:")[1].strip()
                device_control_data.add(EMISSION_START_KEY, emission_start)
                continue

            elif "Excitation Start:" in strp_line:
                excitation_start = strp_line.split("Excitation Start:")[1].strip()
                device_control_data.add(EXCITATION_START_KEY, excitation_start)
                continue

            line_data: list[str] = strp_line.split(",  ")
            for read_datum in line_data:
                splitted_datum = read_datum.split(": ")
                if len(splitted_datum) != 2:
                    continue
                device_control_data.add(splitted_datum[0], splitted_datum[1])

        return device_control_data

    @classmethod
    def _get_step_label(cls, read_line: str, read_mode: str) -> str | None:
        split_line = read_line.strip().split("\t")
        if len(split_line) != 2:
            msg = (
                f"Expected the Read data line {split_line} to contain exactly 2 values."
            )
            raise AllotropeConversionError(msg)
        if split_line[1] != f"{read_mode.title()} Endpoint":
            return split_line[1]

        return None


@dataclass(frozen=True)
class KineticData:
    run_time: str | None
    interval: str | None
    reads: int | None

    @classmethod
    def create(cls, lines: list[str]) -> KineticData | None:
        for line in lines:
            if line.startswith("Start Kinetic\t"):
                return cls._parse_kinetic_section(line)
        return None

    @classmethod
    def _parse_kinetic_section(cls, line: str) -> KineticData:
        split_line = line.strip().split("\t")[1].split(", ")
        run_value = [part for part in split_line if "Runtime" in part]
        if len(run_value) != 1:
            msg = f"Could not find 'Runtime' in Kinetic line: {line}."
            raise AllotropeConversionError(msg)
        interval = [part for part in split_line if "Interval" in part]
        if len(interval) != 1:
            msg = f"Could not find 'Interval' in Kinetic line: {line}."
            raise AllotropeConversionError(msg)
        reads = [part for part in split_line if "Reads" in part]
        if len(reads) != 1:
            msg = f"Could not find 'Reads' in Kinetic line: {line}."
            raise AllotropeConversionError(msg)
        return KineticData(
            run_time=run_value[0].split(" ")[1].strip(),
            interval=interval[0].split(" ")[1].strip(),
            reads=int(reads[0].split(" ")[0].strip()),
        )


@dataclass
class Gen5DataContext:
    header_data: HeaderData
    read_data: list[ReadData]
    kinetic_data: KineticData | None
    results_section: list[str]
    sample_identifiers: dict[str, str]
    concentration_values: dict[str, float | None]
    actual_temperature: float | None
    kinetic_measurements: dict[str, list[float | None]]
    kinetic_elapsed_time: list[float]
    kinetic_errors: dict[str, list[ErrorDocument]]
    wavelength_section: list[str] | None

    @property
    def is_spectral(self) -> bool:
        return bool(self.read_data and self.read_data[0].read_type == ReadType.SPECTRUM)

    @property
    def has_kinetic_data(self) -> bool:
        return self.kinetic_data is not None

    @property
    def has_kinetic_measurements(self) -> bool:
        return bool(self.kinetic_measurements and self.kinetic_elapsed_time)


@dataclass(frozen=True)
class ReadData:
    read_mode: ReadMode
    read_type: ReadType
    measurement_labels: set[str]
    pathlength_correction: str | None
    step_label: str | None
    number_of_averages: float | None
    detector_distance: float | None
    detector_carriage_speed: str | None
    filter_sets: dict[str, FilterSet]
    is_emission: bool = False
    is_excitation: bool = False

    @classmethod
    def create(cls, lines: list[str]) -> list[ReadData]:
        procedure_details = "\n".join(lines)
        read_type = cls.get_read_type(procedure_details)
        if read_type == ReadType.AREASCAN:
            raise AllotropeConversionError(UNSUPPORTED_READ_TYPE_ERROR)
        read_modes = cls.get_read_modes(procedure_details)
        read_sections = list(SectionLinesReader(lines).iter_sections(r"^\s*Read\t"))
        if len(read_modes) != len(read_sections):
            msg = "Expected the number of read modes to match the number of read sections."
            raise AllotropeConversionError(msg)

        read_data_list: list[ReadData] = []
        for read_mode, read_section in zip(read_modes, read_sections, strict=True):
            device_control_data = DeviceControlData.create(
                read_section.lines, read_mode
            )
            measurement_labels, label_aliases = cls._get_measurement_labels(
                device_control_data, read_mode, read_type
            )
            all_labels = {
                alias for aliases in label_aliases.values() for alias in aliases
            }
            number_of_averages = device_control_data.get(MEASUREMENTS_DATA_POINT_KEY)
            read_height = device_control_data.get(READ_HEIGHT_KEY) or ""

            is_emission = bool(
                device_control_data.get_list(FIXED_EXCITATION_KEY)
            ) or "EM Spectrum" in "\n".join(read_section.lines)
            is_excitation = bool(
                device_control_data.get_list(FIXED_EMISSION_KEY)
            ) or "EX Spectrum" in "\n".join(read_section.lines)

            read_data_list.append(
                ReadData(
                    read_mode=read_mode,
                    read_type=read_type,
                    step_label=device_control_data.step_label,
                    measurement_labels=set(measurement_labels) | all_labels,
                    detector_carriage_speed=device_control_data.get(READ_SPEED_KEY),
                    # Absorbance attributes
                    pathlength_correction=device_control_data.get(
                        PATHLENGTH_CORRECTION_KEY
                    ),
                    number_of_averages=try_float_or_none(number_of_averages),
                    # Luminescence attributes
                    detector_distance=try_float_or_none(read_height.split(" ")[0]),
                    # Fluorescence attributes
                    filter_sets=cls._get_filter_sets(
                        measurement_labels,
                        label_aliases,
                        device_control_data,
                        read_mode,
                    ),
                    is_emission=is_emission,
                    is_excitation=is_excitation,
                )
            )
        return read_data_list

    @staticmethod
    def get_read_modes(procedure_details: str) -> list[ReadMode]:
        read_modes = []
        for read_mode in ReadMode:
            # Construct the regex pattern for the current read mode

            pattern = fr"\t{re.escape(read_mode.value)} (?:Endpoint|Spectrum)"
            # Use regex to find all occurrences of the read mode pattern in the procedure details
            matches = re.findall(pattern, procedure_details)
            if matches:
                # Add the read_mode to the list for each match found
                read_modes.extend([read_mode] * len(matches))

        if not read_modes:
            raise AllotropeConversionError(UNSUPPORTED_READ_MODE_ERROR)

        if ReadMode.ALPHALISA in read_modes and ReadMode.FLUORESCENCE in read_modes:
            raise AllotropeConversionError(ALPHALISA_FLUORESCENCE_FOUND)

        # Replace ALPHALISA with FLUORESCENCE
        read_modes = [
            ReadMode.FLUORESCENCE if read_mode == ReadMode.ALPHALISA else read_mode
            for read_mode in read_modes
        ]

        return read_modes

    @staticmethod
    def get_read_type(procedure_details: str) -> ReadType:
        if ReadType.AREASCAN.value in procedure_details:
            return ReadType.AREASCAN
        elif ReadType.SPECTRUM.value in procedure_details:
            return ReadType.SPECTRUM
        # check for this last, because other modes still contain the word "Endpoint"
        elif ReadType.ENDPOINT.value in procedure_details:
            return ReadType.ENDPOINT

        msg = f"Read type not found; expected to find one of {sorted(ReadType._member_names_)}."
        raise AllotropeConversionError(msg)

    @classmethod
    def _get_measurement_labels(
        cls, device_control_data: DeviceControlData, read_mode: str, read_type: str
    ) -> tuple[list[str], dict[str, set[str]]]:
        """Get measurement labels and aliases for the given read mode."""
        step_label = device_control_data.step_label
        label_prefix = f"{step_label}:" if step_label else ""

        if read_mode == ReadMode.ABSORBANCE:
            return cls._get_absorbance_labels(
                label_prefix, device_control_data, read_type
            )
        elif read_mode == ReadMode.FLUORESCENCE:
            return cls._get_fluorescence_labels(label_prefix, device_control_data)
        elif read_mode == ReadMode.LUMINESCENCE:
            return cls._get_luminescence_labels(label_prefix, device_control_data)
        else:
            return [], {}

    @classmethod
    def _get_absorbance_labels(
        cls, label_prefix: str, device_control_data: DeviceControlData, read_type: str
    ) -> tuple[list[str], dict[str, set[str]]]:
        """Get measurement labels for absorbance measurements."""
        measurement_labels = cls._get_absorbance_measurement_labels(
            label_prefix, device_control_data
        )
        if not measurement_labels and read_type == ReadType.SPECTRUM:
            measurement_labels = [f"{label_prefix.rstrip(':')}:Spectrum"]
        return measurement_labels, {}

    @classmethod
    def _get_fluorescence_labels(
        cls, label_prefix: str, device_control_data: DeviceControlData
    ) -> tuple[list[str], dict[str, set[str]]]:
        """Get measurement labels for fluorescence measurements."""
        excitations, emissions = cls._get_excitation_emission_data(device_control_data)

        if not excitations or not emissions:
            return ["Alpha"], {}

        measurement_labels = [
            f"{label_prefix}{excitation},{emission}"
            for excitation, emission in zip(excitations, emissions, strict=True)
        ]

        if device_control_data.is_polarization:
            return cls._handle_polarization_labels(measurement_labels)
        else:
            return cls._create_fluorescence_aliases(
                label_prefix, excitations, emissions
            )

    @classmethod
    def _get_luminescence_labels(
        cls, label_prefix: str, device_control_data: DeviceControlData
    ) -> tuple[list[str], dict[str, set[str]]]:
        """Get measurement labels for luminescence measurements."""
        emissions = (
            device_control_data.get_list(EMISSION_KEY)
            or device_control_data.get_list(FIXED_EMISSION_KEY)
            or device_control_data.get_list(EMISSION_START_KEY)
        )

        measurement_labels = []
        for emission in emissions:
            label = "Lum" if emission in NAN_EMISSION_EXCITATION else emission
            measurement_labels.append(f"{label_prefix}{label}")

        return measurement_labels, {}

    @classmethod
    def _get_excitation_emission_data(
        cls, device_control_data: DeviceControlData
    ) -> tuple[list[str], list[str]]:
        """Extract excitation and emission data from device control data."""
        excitations = (
            device_control_data.get_list(EXCITATION_KEY)
            or device_control_data.get_list(FIXED_EXCITATION_KEY)
            or device_control_data.get_list(EXCITATION_START_KEY)
        )
        emissions = (
            device_control_data.get_list(EMISSION_KEY)
            or device_control_data.get_list(FIXED_EMISSION_KEY)
            or device_control_data.get_list(EMISSION_START_KEY)
        )
        return excitations, emissions

    @classmethod
    def _handle_polarization_labels(
        cls, measurement_labels: list[str]
    ) -> tuple[list[str], dict[str, set[str]]]:
        """Handle polarization measurement labels."""
        if len(measurement_labels) != 2:
            msg = "Expected the Fluorescence Polarization read mode to contain exactly 2 filter sets."
            raise AllotropeConversionError(msg)

        polarization_labels = [
            f"{label} [{light_direction}]"
            for label, light_direction in zip(
                measurement_labels, LIGHT_DIRECTIONS, strict=True
            )
        ]
        return polarization_labels, {}

    @classmethod
    def _create_fluorescence_aliases(
        cls, label_prefix: str, excitations: list[str], emissions: list[str]
    ) -> tuple[list[str], dict[str, set[str]]]:
        """Create fluorescence measurement labels and their aliases."""
        measurement_labels = [
            f"{label_prefix}{excitation},{emission}"
            for excitation, emission in zip(excitations, emissions, strict=True)
        ]

        label_aliases = {
            f"{label_prefix}{excitation},{emission}": {
                f"{label_prefix}{excitation.split('/')[0]},{emission.split('/')[0]}"
            }
            for excitation, emission in zip(excitations, emissions, strict=True)
        }

        return measurement_labels, label_aliases

    @classmethod
    def _get_absorbance_measurement_labels(
        cls, label_prefix: str | None, device_control_data: DeviceControlData
    ) -> list[str]:
        pathlength_correction = device_control_data.get(PATHLENGTH_CORRECTION_KEY)
        measurement_labels = []

        for wavelength in device_control_data.get_list(WAVELENGTHS_KEY):
            label = f"{label_prefix}{wavelength}"
            measurement_labels.append(label)

        if pathlength_correction:
            test, ref = pathlength_correction.split(" / ")
            test_label = f"{label_prefix}{test} [Test]"
            ref_label = f"{label_prefix}{ref} [Ref]"
            measurement_labels.extend([test_label, ref_label])

        return measurement_labels

    @classmethod
    def _get_filter_sets(
        cls,
        measurement_labels: list[str],
        label_aliases: dict[str, set[str]],
        device_control_data: DeviceControlData,
        read_mode: ReadMode,
    ) -> dict[str, FilterSet]:
        if read_mode == ReadMode.ABSORBANCE:
            return {}

        filter_data_extractor = cls._FilterDataExtractor(device_control_data, read_mode)
        filter_data = cls._create_filter_sets_for_labels(
            measurement_labels, filter_data_extractor, device_control_data
        )

        for measurement_label, aliases in label_aliases.items():
            for alias in aliases:
                filter_data[alias] = filter_data[measurement_label]

        return filter_data

    @classmethod
    def _create_filter_sets_for_labels(
        cls,
        measurement_labels: list[str],
        extractor: _FilterDataExtractor,
        device_control_data: DeviceControlData,
    ) -> dict[str, FilterSet]:
        gains = extractor.gains
        if len(measurement_labels) != len(gains):
            msg = f"Expected the number of measurement labels: {measurement_labels} to match the number of gains: {gains}."
            raise AllotropeConversionError(msg)

        filter_data = {}
        for idx, label in enumerate(measurement_labels):
            filter_set = cls._create_single_filter_set(
                idx, extractor, device_control_data
            )
            filter_data[label] = filter_set

        return filter_data

    @classmethod
    def _create_single_filter_set(
        cls,
        idx: int,
        extractor: _FilterDataExtractor,
        device_control_data: DeviceControlData,
    ) -> FilterSet:
        excitation = cls._clean_wavelength_value(extractor.get_excitation_at_index(idx))
        emission = cls._clean_wavelength_value(extractor.get_emission_at_index(idx))
        mirror = extractor.get_mirror_at_index(idx)
        optics = extractor.get_optics_at_index(idx)
        gain = extractor.gains[idx]

        light_direction = (
            LIGHT_DIRECTIONS[idx] if device_control_data.is_polarization else None
        )

        return FilterSet(
            emission=emission,
            gain=gain,
            excitation=excitation,
            mirror=mirror,
            optics=optics,
            light_direction=light_direction,
        )

    @staticmethod
    def _clean_wavelength_value(wavelength: str | None) -> str | None:
        if wavelength and " nm" in wavelength:
            return wavelength.split(" nm")[0]
        return wavelength

    class _FilterDataExtractor:
        def __init__(self, device_control_data: DeviceControlData, read_mode: ReadMode):
            self.device_control_data = device_control_data
            self.read_mode = read_mode

            self.emissions = device_control_data.get_list(EMISSION_KEY)
            self.excitations = device_control_data.get_list(EXCITATION_KEY)
            self.fixed_emissions = device_control_data.get_list(FIXED_EMISSION_KEY)
            self.fixed_excitations = device_control_data.get_list(FIXED_EXCITATION_KEY)
            self.emission_start = device_control_data.get_list(EMISSION_START_KEY)
            self.excitation_start = device_control_data.get_list(EXCITATION_START_KEY)
            self.mirrors = device_control_data.get_list(MIRROR_KEY)
            self.optics = device_control_data.get_list(OPTICS_KEY)
            self.gains = device_control_data.get_list(GAIN_KEY)

        def get_excitation_at_index(self, idx: int) -> str | None:
            return (
                self._safe_get(self.excitations, idx)
                or self._safe_get(self.fixed_excitations, idx)
                or self._safe_get(self.excitation_start, idx)
            )

        def get_emission_at_index(self, idx: int) -> str | None:
            return (
                self._safe_get(self.emissions, idx)
                or self._safe_get(self.fixed_emissions, idx)
                or self._safe_get(self.emission_start, idx)
            )

        def get_mirror_at_index(self, idx: int) -> str | None:
            if self.mirrors and self.read_mode == ReadMode.FLUORESCENCE:
                return self._safe_get(self.mirrors, idx)
            return None

        def get_optics_at_index(self, idx: int) -> str | None:
            return self._safe_get(self.optics, idx)

        @staticmethod
        def _safe_get(data_list: list[str], idx: int) -> str | None:
            return data_list[idx] if data_list and idx < len(data_list) else None


def get_concentrations(layout_lines: list[str] | None) -> dict[str, float | None]:
    """Extract concentration/dilution values from the Layout section.

    Returns a dictionary mapping well positions to their concentration/dilution values.
    """
    if not layout_lines:
        return {}

    # Create dataframe from tabular data and forward fill empty values in index
    data = read_csv(StringIO("\n".join(layout_lines[1:])), sep="\t")
    data = data.set_index(data.index.to_series().ffill(axis="index").values)

    concentrations = {}
    for row_name, row in data.iterrows():
        label = row.iloc[-1]
        if label == "Conc/Dil":
            for col_index, col in enumerate(row.iloc[:-1]):
                well_pos = f"{row_name}{col_index + 1}"
                # Convert to float if possible
                if not pd.isna(col):
                    concentration_value = try_float_or_none(col)
                    concentrations[well_pos] = concentration_value
                else:
                    concentrations[well_pos] = None

    return concentrations


def get_identifiers(layout_lines: list[str] | None) -> dict[str, str]:
    if not layout_lines:
        return {}
    # Create dataframe from tabular data and forward fill empty values in index
    data = read_csv(StringIO("\n".join(layout_lines[1:])), sep="\t")
    data = data.set_index(data.index.to_series().ffill(axis="index").values)

    identifiers = {}
    for row_name, row in data.iterrows():
        label = row.iloc[-1]
        for col_index, col in enumerate(row.iloc[:-1]):
            well_pos = f"{row_name}{col_index + 1}"
            # Prefer Name to Well ID
            if not pd.isna(col) and (
                label == "Name" or label == "Well ID" and well_pos not in identifiers
            ):
                identifiers[well_pos] = col
    return identifiers


def get_temperature(actual_temperature_lines: list[str] | None) -> float | None:
    if not actual_temperature_lines:
        return None
    if len(actual_temperature_lines) != 1:
        msg = f"Expected the Temperature section '{actual_temperature_lines}' to contain exactly 1 line."
        raise AllotropeConversionError(msg)

    return try_float(
        actual_temperature_lines[0].strip().split("\t")[-1], "Actual Temperature"
    )


def get_kinetic_measurements(
    kinetic_lines: list[str] | None,
) -> tuple[
    dict[str, list[float | None]], list[float], dict[str, list[ErrorDocument]]
] | None:
    if not kinetic_lines:
        return None
    data = (
        read_csv(StringIO("\n".join(kinetic_lines)), sep="\t", index_col=0)
        .dropna(axis="columns", how="all")
        .dropna(axis="index", how="all")
    )

    kinetic_measurements: defaultdict[str, list[float | None]] = defaultdict(
        list[float | None]
    )
    kinetic_elapsed_time: list[float] = data.index.map(
        lambda val: _convert_time_to_seconds(str(val))
    ).to_list()

    error_documents: dict[str, list[ErrorDocument]] = {}

    for col_name, column in data.items():
        # Skip temperature columns (e.g., "T∞ 600" or "T° 600")
        col_name_str = str(col_name)
        if col_name_str.startswith("T") and any(
            char in col_name_str for char in ["∞", "°", "\u221e", "\u00b0"]
        ):
            continue

        well_values: list[float | None] = []
        for idx, value in enumerate(column):
            # Handle empty values
            if pd.isna(value):
                well_values.append(None)
                continue

            try:
                float_value = float(value)
                well_values.append(float_value)
            except (ValueError, TypeError):
                if value is not None:
                    str_value = (
                        str(value).strip()
                        if not isinstance(value, str)
                        else value.strip()
                    )
                    if str_value:  # Only create error for non-empty values
                        time_point = (
                            kinetic_elapsed_time[idx]
                            if idx < len(kinetic_elapsed_time)
                            else None
                        )

                        if str(col_name) not in error_documents:
                            error_documents[str(col_name)] = []

                        error_documents[str(col_name)].append(
                            ErrorDocument(
                                error=str_value,
                                error_feature=f"{time_point}s",
                            )
                        )
                # Use NEGATIVE_ZERO for non-numeric values
                well_values.append(NEGATIVE_ZERO)

        kinetic_measurements[str(col_name)] = well_values

    return dict(kinetic_measurements.items()), kinetic_elapsed_time, error_documents


@dataclass(frozen=True)
class MeasurementData:
    identifier: str
    value: float | None
    label: str


def _get_label(row: pd.Series[Any], measurement_labels: set[str]) -> tuple[str, bool]:
    raw_label = row.iloc[-1]
    label_str = str(raw_label)

    # Strip "Read X:" prefix if present (e.g., "Read 1:450,490" -> "450,490")
    if label_str.startswith("Read ") and ":" in label_str:
        label_str = label_str.split(":", 1)[1]

    if label_str in measurement_labels:
        return label_str, True
    if isinstance(raw_label, float) and str(int(raw_label)) in measurement_labels:
        return str(int(raw_label)), True
    return label_str, False


def create_results(
    result_lines: list[str],
    header_data: HeaderData,
    read_data: list[ReadData],
    sample_identifiers: dict[str, str],
    actual_temperature: float | None,
    concentration_values: dict[str, float | None] | None = None,
) -> tuple[list[MeasurementGroup], list[CalculatedDocument]]:
    if not result_lines:
        raise AllotropeConversionError(NO_MEASUREMENTS_ERROR)

    if result_lines[0].strip() != "Results":
        msg = f"Expected the first line of the results section '{result_lines[0]}' to be 'Results'."
        raise AllotropeConversionError(msg)

    # Create dataframe from tabular data and forward fill empty values in index
    data = read_csv(StringIO("\n".join(result_lines[1:])), sep="\t")
    data = data.set_index(data.index.to_series().ffill(axis="index").values)
    well_to_measurements: defaultdict[str, list[MeasurementData]] = defaultdict(
        list[MeasurementData]
    )
    calculated_data: defaultdict[str, list[tuple[str, float]]] = defaultdict(
        list[tuple[str, float]]
    )
    measurement_labels = {
        label for r_data in read_data for label in r_data.measurement_labels
    }
    error_documents_per_well: defaultdict[str, list[ErrorDocument]] = defaultdict(
        list[ErrorDocument]
    )

    for row_name, row in data.iterrows():
        label, is_measurement = _get_label(row, measurement_labels)

        for col_index, value in enumerate(row.iloc[:-1]):
            well_pos = f"{row_name}{col_index + 1}"
            well_value = try_float_or_none(value)
            # skip empty values
            if isinstance(well_value, float) and math.isnan(well_value):
                continue

            # Report error documents for NaN values
            if well_value is None:
                error_documents_per_well[well_pos].append(
                    ErrorDocument(
                        error=value,
                        # TODO Add support for multiple read modes
                        error_feature=(
                            read_data[0].read_mode.lower() if is_measurement else label
                        ),
                    )
                )
            if is_measurement:
                well_to_measurements[well_pos].append(
                    MeasurementData(
                        random_uuid_str(),
                        NEGATIVE_ZERO if well_value is None else well_value,
                        label,
                    )
                )
            elif well_value is not None:
                calculated_data[well_pos].append((label, well_value))

    # Calculate plate well count from data dimensions and round to nearest valid count
    calculated_well_count = len(set(data.index.tolist())) * len(
        set(data.columns[1:].tolist())
    )
    plate_well_count = (
        round_to_nearest_well_count(calculated_well_count) or calculated_well_count
    )

    groups = [
        MeasurementGroup(
            measurement_time=header_data.datetime,
            plate_well_count=plate_well_count,
            measurements=[
                _create_measurement(
                    measurement,
                    well_position,
                    header_data,
                    get_read_data_from_measurement(measurement, read_data),
                    sample_identifiers.get(well_position),
                    actual_temperature,
                    error_documents=error_documents_per_well.get(well_position),
                    concentration_value=concentration_values.get(well_position)
                    if concentration_values
                    else None,
                )
                for measurement in measurements
            ],
        )
        for well_position, measurements in well_to_measurements.items()
    ]

    calculated_data_items = []
    for well_position, well_calculated_data in calculated_data.items():
        if well_position not in well_to_measurements:
            continue  # Skip wells without measurements
        for label, value in well_calculated_data:
            data_sources = [
                DataSource(
                    reference=Referenceable(measurement.identifier),
                    feature=item.read_mode.value.lower(),
                )
                for measurement in _get_sources(
                    label, well_to_measurements[well_position]
                )
                for item in read_data
            ]
            # Only create calculated document if we have data sources
            if data_sources:
                calculated_data_items.append(
                    CalculatedDocument(
                        uuid=random_uuid_str(),
                        data_sources=data_sources,
                        unit=UNITLESS,
                        name=label,
                        value=value,
                    )
                )

    return groups, calculated_data_items


def create_kinetic_results(
    result_lines: list[str],
    header_data: HeaderData,
    read_data: list[ReadData],
    sample_identifiers: dict[str, str],
    actual_temperature: float | None,
    kinetic_data: KineticData,
    kinetic_measurements: dict[str, list[float | None]],
    kinetic_elapsed_time: list[float],
    kinetic_errors: dict[str, list[ErrorDocument]] | None = None,
    concentration_values: dict[str, float | None] | None = None,
) -> tuple[list[MeasurementGroup], list[CalculatedDocument]]:
    calculated_data: defaultdict[str, list[tuple[str, float]]] = defaultdict(
        list[tuple[str, float]]
    )
    measurement_labels = [
        label for r_data in read_data for label in r_data.measurement_labels
    ]

    error_documents_per_well: defaultdict[str, list[ErrorDocument]] = defaultdict(list)

    if kinetic_errors:
        for well, errors in kinetic_errors.items():
            error_documents_per_well[well].extend(errors)

    # Parse calculated data from results section if present
    if result_lines and result_lines[0].strip() == "Results":
        # Create dataframe from tabular data and forward fill empty values in index
        data = read_csv(StringIO("\n".join(result_lines[1:])), sep="\t")
        data = data.set_index(data.index.to_series().ffill(axis="index").values)

        for row_name, row in data.iterrows():
            label = row.iloc[-1]
            for col_index, value in enumerate(row.iloc[:-1]):
                well_pos = f"{row_name}{col_index + 1}"
                well_value = try_non_nan_float_or_none(value)
                # TODO: Report error documents for NaN values
                if not well_value:
                    continue
                calculated_data[well_pos].append((label, well_value))

    # Get all well positions from kinetic measurements (primary) or calculated data
    # Preserve order from kinetic_measurements, then add any wells only in calculated_data
    well_positions = list(kinetic_measurements.keys())

    groups = [
        MeasurementGroup(
            measurement_time=header_data.datetime,
            plate_well_count=get_well_count_by_well_ids(well_locations=well_positions),
            measurements=[
                _create_measurement(
                    measurement := MeasurementData(
                        random_uuid_str(), None, measurement_labels[0]
                    ),
                    well_position,
                    header_data,
                    get_read_data_from_measurement(measurement, read_data),
                    sample_identifiers.get(well_position),
                    actual_temperature,
                    kinetic_data,
                    kinetic_measurements[well_position],
                    kinetic_elapsed_time,
                    error_documents_per_well.get(well_position, []),
                    concentration_value=concentration_values.get(well_position)
                    if concentration_values
                    else None,
                )
            ],
        )
        for well_position in well_positions
    ]

    groups_by_well_position = {
        group.measurements[0].location_identifier: group for group in groups
    }

    calculated_data_items = [
        CalculatedDocument(
            uuid=random_uuid_str(),
            data_sources=[
                DataSource(
                    reference=Referenceable(
                        groups_by_well_position[well_position]
                        .measurements[0]
                        .identifier
                    ),
                    # TODO: Add support for multiple kinetic sections
                    feature=DATA_SOURCE_FEATURE_VALUES[read_data[0].read_mode],
                )
            ],
            unit=UNITLESS,
            name=label,
            value=value,
        )
        for well_position, well_calculated_data in calculated_data.items()
        for label, value in well_calculated_data
    ]

    return groups, calculated_data_items


def get_read_data_from_measurement(
    measurement: MeasurementData, read_data_list: list[ReadData]
) -> ReadData:
    for read_data in read_data_list:
        if _is_label_in_measurement_labels(
            measurement.label, read_data.measurement_labels
        ):
            return read_data

    raise AllotropeConversionError(
        READ_DATA_MEASUREMENT_ERROR.format(measurement.label)
    )


def _get_sources(
    calculated_data_label: str, measurements: list[MeasurementData]
) -> list[MeasurementData]:
    # Pathlength is a special case, its sources are always determined
    # by the pathlength correction setting
    if calculated_data_label.split(":")[-1] == "Pathlength":
        pathlength_suffixes = ["[Test]", "[Ref]"]
        sources = [
            measurement
            for measurement in measurements
            if measurement.label.split(" ")[-1] in pathlength_suffixes
        ]
    else:
        sources = [
            measurement
            for measurement in measurements
            if measurement.label in calculated_data_label
        ]
    # if there are no matches in the measurement labels, use all measurements as source
    return sources or measurements


def create_metadata(header_data: HeaderData) -> Metadata:
    asm_file_identifier = Path(header_data.file_name).with_suffix(".json")
    return Metadata(
        device_identifier=NOT_APPLICABLE,
        model_number=header_data.model_number or NOT_APPLICABLE,
        equipment_serial_number=header_data.equipment_serial_number,
        software_name=DEFAULT_SOFTWARE_NAME,
        software_version=header_data.software_version,
        file_name=header_data.file_name,
        asm_file_identifier=asm_file_identifier.name,
        data_system_instance_id=NOT_APPLICABLE,
        unc_path=header_data.unc_path,
    )


def create_calculated_data_documents(
    results_section: list[str] | None,
    read_data: ReadData,
    measurements: list[Measurement],
) -> tuple[list[CalculatedDocument], dict[str, list[ErrorDocument]]]:
    calculated_data = []
    error_documents_by_well: defaultdict[str, list[ErrorDocument]] = defaultdict(list)
    if results_section:
        if results_section[0].strip() != "Results":
            return [], {}

        data = read_csv(StringIO("\n".join(results_section[1:])), sep="\t")

        calculated_data_by_well = defaultdict(list)
        for row_name, row in data.iterrows():
            label = str(row.iloc[-1])
            # Skip rows containing measurement data
            if label in read_data.measurement_labels:
                continue

            for col_index, value in enumerate(row.iloc[:-1]):
                well_pos = f"{row_name}{col_index + 1}"
                well_value = try_float_or_none(value)

                # Handle numeric values
                if well_value is not None and not math.isnan(well_value):
                    calculated_data_by_well[well_pos].append((label, well_value))
                    continue

                # Handle empty cells
                if pd.isna(value) or value is None or str(value).strip() == "":
                    continue

                # Handle non-numeric error values
                error_documents_by_well[well_pos].append(
                    ErrorDocument(
                        error=str(value).strip(),
                        error_feature=label,
                    )
                )
                calculated_data_by_well[well_pos].append((label, NEGATIVE_ZERO))

        well_to_measurement_id = {
            measurement.location_identifier: measurement.identifier
            for measurement in measurements
        }

        for well_position, well_calculated_data in calculated_data_by_well.items():
            for label, value in well_calculated_data:
                calculated_data.append(
                    CalculatedDocument(
                        uuid=random_uuid_str(),
                        data_sources=[
                            DataSource(
                                reference=Referenceable(
                                    well_to_measurement_id[well_position]
                                ),
                                feature=read_data.read_mode.value.lower(),
                            )
                        ],
                        unit=UNITLESS,
                        name=label,
                        value=value,
                    )
                )
    return calculated_data, error_documents_by_well


def create_spectrum_results(
    header_data: HeaderData,
    read_data_list: list[ReadData],
    wavelengths_sections: list[str] | None,
    sample_identifiers: dict[str, str],
    actual_temperature: float | None,
    results_section: list[str] | None = None,
    concentration_values: dict[str, float | None] | None = None,
) -> tuple[list[MeasurementGroup], list[CalculatedDocument]]:
    if not wavelengths_sections:
        return [], []

    data = read_csv(StringIO("\n".join(wavelengths_sections)), sep="\t")

    try:
        wavelengths = data["Wavelength"].astype(float).tolist()
    except (ValueError, KeyError):
        return [], []

    wells_data = {}
    error_documents_by_well = {}

    for column in data.columns:
        if column == "Wavelength":
            continue

        well_values = []
        errors = []
        for idx, value in enumerate(data[column]):
            try:
                float_value = float(value)
                if math.isnan(float_value):
                    continue
                well_values.append(float_value)
            except (ValueError, TypeError):
                original_value = str(value).strip()
                if original_value:
                    wavelength = wavelengths[idx] if idx < len(wavelengths) else None
                    errors.append(
                        ErrorDocument(
                            error=original_value,
                            error_feature=f"{wavelength}nm",
                        )
                    )
                well_values.append(NEGATIVE_ZERO)

        # Skip empty wells
        if not well_values:
            continue

        wells_data[column] = well_values
        if errors:
            error_documents_by_well[column] = errors

    if not wells_data:
        return [], []

    read_data = read_data_list[0]
    if not read_data:
        return [], []

    measurements: list[Measurement] = []
    if read_data.read_mode == ReadMode.ABSORBANCE:
        measurement_type = MeasurementType.ULTRAVIOLET_ABSORBANCE_CUBE_SPECTRUM
    elif read_data.read_mode == ReadMode.FLUORESCENCE and read_data.is_emission:
        measurement_type = MeasurementType.EMISSION_FLUORESCENCE_CUBE_SPECTRUM
    elif read_data.read_mode == ReadMode.FLUORESCENCE and read_data.is_excitation:
        measurement_type = MeasurementType.EXCITATION_FLUORESCENCE_CUBE_SPECTRUM
    elif read_data.read_mode == ReadMode.FLUORESCENCE:
        # Default to emission if not specified
        measurement_type = MeasurementType.EMISSION_FLUORESCENCE_CUBE_SPECTRUM
    elif read_data.read_mode == ReadMode.LUMINESCENCE and read_data.is_emission:
        measurement_type = MeasurementType.EMISSION_LUMINESCENCE_CUBE_SPECTRUM
    elif read_data.read_mode == ReadMode.LUMINESCENCE and read_data.is_excitation:
        measurement_type = MeasurementType.EXCITATION_LUMINESCENCE_CUBE_SPECTRUM
    elif read_data.read_mode == ReadMode.LUMINESCENCE:
        # Default to emission if not specified
        measurement_type = MeasurementType.EMISSION_LUMINESCENCE_CUBE_SPECTRUM
    else:
        msg = f"Unsupported read mode: {read_data.read_mode}"
        raise AllotropeConversionError(msg)

    # Extract wavelength settings from filter sets
    filter_set = None
    if read_data.filter_sets:
        filter_set = next(iter(read_data.filter_sets.values()))

    detector_wavelength_setting = (
        filter_set.detector_wavelength_setting
        if filter_set and read_data.is_excitation
        else None
    )
    detector_bandwidth_setting = (
        filter_set.detector_bandwidth_setting if filter_set else None
    )
    excitation_wavelength_setting = (
        filter_set.excitation_wavelength_setting
        if filter_set and read_data.is_emission
        else None
    )
    excitation_bandwidth_setting = (
        filter_set.excitation_bandwidth_setting if filter_set else None
    )

    # Special case for luminescence spectrum
    if read_data.read_mode == ReadMode.LUMINESCENCE:
        excitation_wavelength_setting = 480.0
        excitation_bandwidth_setting = 20.0

    for well_position, well_absorbance_data in wells_data.items():
        if measurement_type == MeasurementType.ULTRAVIOLET_ABSORBANCE_CUBE_SPECTRUM:
            label = "absorbance-spectrum"
        elif measurement_type in (
            MeasurementType.EMISSION_FLUORESCENCE_CUBE_SPECTRUM,
            MeasurementType.EXCITATION_FLUORESCENCE_CUBE_SPECTRUM,
        ):
            label = "fluorescence-spectrum"
        elif measurement_type in (
            MeasurementType.EMISSION_LUMINESCENCE_CUBE_SPECTRUM,
            MeasurementType.EXCITATION_LUMINESCENCE_CUBE_SPECTRUM,
        ):
            label = "luminescence-spectrum"
        else:
            label = "spectrum"

        spectrum_data_cube = DataCube(
            label=label,
            structure_dimensions=[
                DataCubeComponent(
                    concept="wavelength",
                    type_=FieldComponentDatatype.double,
                    unit=Nanometer.unit,
                )
            ],
            structure_measures=[
                DataCubeComponent(
                    concept="absorbance",
                    type_=FieldComponentDatatype.double,
                    unit=MilliAbsorbanceUnit.unit,
                )
            ],
            dimensions=[wavelengths],
            measures=[well_absorbance_data],
        )

        measurement = Measurement(
            type_=measurement_type,
            device_type=DEVICE_TYPE,
            identifier=random_uuid_str(),
            sample_identifier=sample_identifiers.get(well_position)
            or f"{header_data.well_plate_identifier} {well_position}",
            location_identifier=well_position,
            well_plate_identifier=header_data.well_plate_identifier,
            detection_type=read_data.read_mode.value,
            detector_wavelength_setting=detector_wavelength_setting,
            detector_bandwidth_setting=detector_bandwidth_setting,
            excitation_wavelength_setting=excitation_wavelength_setting,
            excitation_bandwidth_setting=excitation_bandwidth_setting,
            compartment_temperature=actual_temperature,
            spectrum_data_cube=spectrum_data_cube,
            number_of_averages=read_data.number_of_averages,
            detector_carriage_speed=read_data.detector_carriage_speed,
            detector_distance_setting=read_data.detector_distance,
            scan_position_setting=filter_set.scan_position_setting
            if filter_set
            else None,
            detector_gain_setting=filter_set.gain if filter_set else None,
            error_document=error_documents_by_well.get(well_position),
            sample_custom_info={
                "Conc/Dil": concentration_values.get(well_position)
                if concentration_values
                else None,
            },
            analytical_method_identifier=header_data.protocol_file_path
            if header_data.protocol_file_path
            else None,
            experimental_data_identifier=header_data.experiment_file_path
            if header_data.experiment_file_path
            else None,
        )

        measurements.append(measurement)

    calculated_data, calculated_data_errors = create_calculated_data_documents(
        results_section, read_data_list[0], measurements
    )

    # Add calculated data errors to measurement error documents
    for well_position, errors in calculated_data_errors.items():
        for measurement in measurements:
            if measurement.location_identifier == well_position:
                if measurement.error_document:
                    measurement.error_document.extend(errors)
                else:
                    measurement.error_document = errors

    # Calculate plate_well_count from all well positions in header (including empty wells)
    plate_well_count = get_well_count_by_well_ids(
        well_locations=[col for col in data.columns if col != "Wavelength"]
    )

    measurement_groups = [
        MeasurementGroup(
            measurement_time=header_data.datetime,
            plate_well_count=plate_well_count,
            measurements=[measurement],
        )
        for measurement in measurements
    ]

    return measurement_groups, calculated_data


def _is_label_in_measurement_labels(label: str, measurement_labels: set[str]) -> bool:
    if not measurement_labels:
        return False
    if label in measurement_labels:
        return True
    # TODO Improve this logic to support multiple kinetic sections
    match = re.search(r"\[(\d+)]", label)
    if match and match.group()[1:-1] in measurement_labels:
        return True
    return False


def _create_measurement(
    measurement: MeasurementData,
    well_position: str,
    header_data: HeaderData,
    read_data: ReadData,
    sample_identifier: str | None,
    actual_temperature: float | None,
    kinetic_data: KineticData | None = None,
    kinetic_measurements: list[float | None] | None = None,
    kinetic_elapsed_time: list[float] | None = None,
    error_documents: list[ErrorDocument] | None = None,
    concentration_value: float | None = None,
) -> Measurement:
    if read_data.read_mode == ReadMode.ABSORBANCE and not kinetic_data:
        measurement_type = MeasurementType.ULTRAVIOLET_ABSORBANCE
    elif read_data.read_mode == ReadMode.FLUORESCENCE and not kinetic_data:
        measurement_type = MeasurementType.FLUORESCENCE
    elif read_data.read_mode == ReadMode.LUMINESCENCE and not kinetic_data:
        measurement_type = MeasurementType.LUMINESCENCE
    elif read_data.read_mode == ReadMode.ABSORBANCE and kinetic_data:
        measurement_type = MeasurementType.ULTRAVIOLET_ABSORBANCE_CUBE_DETECTOR
    elif read_data.read_mode == ReadMode.FLUORESCENCE and kinetic_data:
        measurement_type = MeasurementType.FLUORESCENCE_CUBE_DETECTOR
    elif read_data.read_mode == ReadMode.LUMINESCENCE and kinetic_data:
        measurement_type = MeasurementType.LUMINESCENCE_CUBE_DETECTOR

    if measurement_type in [
        MeasurementType.ULTRAVIOLET_ABSORBANCE,
        MeasurementType.ULTRAVIOLET_ABSORBANCE_CUBE_DETECTOR,
    ]:
        filter_data = None
        detector_wavelength_setting = (
            float(measurement.label.split(":")[-1].split(" ")[0])
        ) or None
    else:
        filter_data = read_data.filter_sets[measurement.label]
        detector_wavelength_setting = filter_data.detector_wavelength_setting

    return Measurement(
        type_=measurement_type,
        device_type=DEVICE_TYPE,
        identifier=measurement.identifier,
        sample_identifier=sample_identifier
        or f"{header_data.well_plate_identifier} {well_position}",
        location_identifier=well_position,
        well_plate_identifier=header_data.well_plate_identifier,
        detection_type=read_data.read_mode.value,
        detector_wavelength_setting=detector_wavelength_setting,
        detector_bandwidth_setting=(
            filter_data.detector_bandwidth_setting if filter_data else None
        ),
        excitation_wavelength_setting=(
            filter_data.excitation_wavelength_setting if filter_data else None
        ),
        excitation_bandwidth_setting=(
            filter_data.excitation_bandwidth_setting if filter_data else None
        ),
        wavelength_filter_cutoff_setting=(
            filter_data.wavelength_filter_cutoff_setting if filter_data else None
        ),
        detector_distance_setting=read_data.detector_distance,
        scan_position_setting=(
            filter_data.scan_position_setting if filter_data else None
        ),
        detector_gain_setting=filter_data.gain if filter_data else None,
        number_of_averages=read_data.number_of_averages,
        detector_carriage_speed=read_data.detector_carriage_speed,
        absorbance=(
            measurement.value
            if measurement_type == MeasurementType.ULTRAVIOLET_ABSORBANCE
            and not kinetic_data
            else None
        ),
        fluorescence=(
            measurement.value
            if measurement_type == MeasurementType.FLUORESCENCE and not kinetic_data
            else None
        ),
        luminescence=(
            measurement.value
            if measurement_type == MeasurementType.LUMINESCENCE and not kinetic_data
            else None
        ),
        compartment_temperature=actual_temperature,
        total_measurement_time_setting=(
            _convert_time_to_seconds(kinetic_data.run_time) if kinetic_data else None
        ),
        read_interval_setting=(
            _convert_time_to_seconds(kinetic_data.interval) if kinetic_data else None
        ),
        number_of_scans_setting=kinetic_data.reads if kinetic_data else None,
        profile_data_cube=(
            DataCube(
                label=str(detector_wavelength_setting),
                structure_dimensions=[
                    DataCubeComponent(
                        concept=ELAPSED_TIME,
                        type_=FieldComponentDatatype.double,
                        unit=SECONDS,
                    )
                ],
                structure_measures=[
                    DataCubeComponent(
                        concept=read_data.read_mode.lower(),
                        type_=FieldComponentDatatype.double,
                        unit=READ_MODE_UNITS.get(read_data.read_mode, UNITLESS),
                    )
                ],
                dimensions=[kinetic_elapsed_time],
                measures=[kinetic_measurements],
            )
            if kinetic_elapsed_time and kinetic_measurements
            else None
        ),
        error_document=error_documents,
        device_control_custom_info={
            "Reading Type": header_data.additional_data.pop("Reading Type", None),
            "Direction of emitted light": (
                filter_data.light_direction if filter_data else None
            ),
        },
        sample_custom_info={
            "Plate Number": header_data.additional_data.pop("Plate Number", None),
            "Conc/Dil": concentration_value,
        },
        measurement_custom_info=header_data.additional_data,
        analytical_method_identifier=header_data.protocol_file_path
        if header_data.protocol_file_path
        else None,
        experimental_data_identifier=header_data.experiment_file_path
        if header_data.experiment_file_path
        else None,
    )


def _convert_time_to_seconds(time_str: str | None) -> float:
    if not time_str:
        return 0.0
    try:
        hours, minutes, seconds = map(float, time_str.split(":"))
        return hours * 3600 + minutes * 60 + seconds

    except ValueError:
        msg = f"Invalid time string: '{time_str}'."
        raise AllotropeConversionError(msg) from None


class ResultProcessor(ABC):
    """Abstract base class for processing different types of Gen5 results."""

    @classmethod
    @abstractmethod
    def can_process(cls, context: Gen5DataContext) -> bool:
        """Check if this processor can handle the given context."""
        pass

    @abstractmethod
    def process(
        self, context: Gen5DataContext
    ) -> tuple[list[MeasurementGroup], list[CalculatedDocument]]:
        """Process the context and return measurement groups and calculated data."""
        pass


class SpectralResultProcessor(ResultProcessor):
    """Processor for spectral measurements with wavelength data."""

    @classmethod
    def can_process(cls, context: Gen5DataContext) -> bool:
        return context.is_spectral and context.wavelength_section is not None

    def process(
        self, context: Gen5DataContext
    ) -> tuple[list[MeasurementGroup], list[CalculatedDocument]]:
        return create_spectrum_results(
            context.header_data,
            read_data_list=context.read_data,
            wavelengths_sections=context.wavelength_section,
            sample_identifiers=context.sample_identifiers,
            actual_temperature=context.actual_temperature,
            results_section=context.results_section,
            concentration_values=context.concentration_values,
        )


class KineticResultProcessor(ResultProcessor):
    """Processor for kinetic measurements with time series data."""

    @classmethod
    def can_process(cls, context: Gen5DataContext) -> bool:
        if not context.has_kinetic_data:
            return False

        if not context.has_kinetic_measurements:
            msg = "Kinetic data is present in the file but no kinetic measurements data is found."
            raise AllotropeConversionError(msg)

        return True

    def process(
        self, context: Gen5DataContext
    ) -> tuple[list[MeasurementGroup], list[CalculatedDocument]]:
        return create_kinetic_results(
            context.results_section,
            context.header_data,
            context.read_data,
            context.sample_identifiers,
            context.actual_temperature,
            cast(KineticData, context.kinetic_data),
            context.kinetic_measurements,
            context.kinetic_elapsed_time,
            context.kinetic_errors,
            context.concentration_values,
        )


class StandardResultProcessor(ResultProcessor):
    """Processor for standard endpoint measurements."""

    @classmethod
    def can_process(cls, _: Gen5DataContext) -> bool:
        return True

    def process(
        self, context: Gen5DataContext
    ) -> tuple[list[MeasurementGroup], list[CalculatedDocument]]:
        return create_results(
            context.results_section,
            context.header_data,
            context.read_data,
            context.sample_identifiers,
            context.actual_temperature,
            context.concentration_values,
        )


def get_processor(
    context: Gen5DataContext,
) -> SpectralResultProcessor | KineticResultProcessor | StandardResultProcessor:
    """Factory function to get the appropriate processor for the given context."""
    if SpectralResultProcessor.can_process(context):
        return SpectralResultProcessor()
    if KineticResultProcessor.can_process(context):
        return KineticResultProcessor()
    if StandardResultProcessor.can_process(context):
        return StandardResultProcessor()

    msg = "No suitable processor found for the data."
    raise AllotropeConversionError(msg)


def merge_measurement_groups(
    all_groups: list[list[MeasurementGroup]],
) -> list[MeasurementGroup]:
    """Merge multiple lists of measurement groups by well position.

    Each list of measurement groups represents results from a different read/filter set.
    This function combines them so that each well has all its measurements in a single group.
    """
    # Group by well position
    groups_by_well: dict[str, list[MeasurementGroup]] = defaultdict(list)
    for group_list in all_groups:
        for group in group_list:
            # Get the well position from the first measurement in the group
            if group.measurements:
                well_pos = group.measurements[0].location_identifier
                groups_by_well[well_pos].append(group)

    # Merge measurements for each well
    merged_groups = []
    for well_pos in groups_by_well.keys():
        groups = groups_by_well[well_pos]
        # Take metadata from the first group
        base_group = groups[0]
        # Collect all measurements from all groups for this well
        all_measurements = []
        for group in groups:
            all_measurements.extend(group.measurements)

        # Create merged group
        merged_group = MeasurementGroup(
            measurement_time=base_group.measurement_time,
            plate_well_count=base_group.plate_well_count,
            measurements=all_measurements,
        )
        merged_groups.append(merged_group)

    return merged_groups


def process_all_reads(
    reader: Any, context: Gen5DataContext
) -> tuple[list[MeasurementGroup], list[CalculatedDocument]]:
    """Process all ReadData entries and return merged measurement groups and calculated data.

    Args:
        reader: The AgilentGen5Reader instance
        context: The main Gen5DataContext with all data

    Returns:
        Tuple of (merged measurement groups, calculated data documents)
    """
    # Process each ReadData entry separately and collect results
    all_measurement_groups: list[list[MeasurementGroup]] = []
    all_calculated_data: list[CalculatedDocument] = []

    # Track which reads use the combined results section to avoid processing it multiple times
    combined_results_read_indices: list[int] = []

    for read_index, read_data in enumerate(context.read_data, start=1):
        # Create a test sub-context to check if this is kinetic
        test_context = reader.create_read_context(context, read_data, read_index)
        is_kinetic = test_context.has_kinetic_measurements

        if is_kinetic and len(read_data.measurement_labels) > 1:
            # Split kinetic read into separate sub-reads per filter set
            # Each label gets its own Time section processed separately
            # Sort labels consistently to ensure stable output
            sorted_labels = sorted(
                read_data.measurement_labels, key=lambda x: (len(x), x)
            )
            for label in sorted_labels:
                sub_context = reader.create_read_context(
                    context, read_data, read_index, specific_label=label
                )
                processor = get_processor(sub_context)
                measurement_groups, calculated_data = processor.process(sub_context)
                all_measurement_groups.append(measurement_groups)
                all_calculated_data.extend(calculated_data)
        else:
            # Check if this read uses the combined results section
            uses_combined_results = (
                not test_context.has_kinetic_measurements
                and test_context.results_section
                and reader.get_results_section() == test_context.results_section
            )

            if uses_combined_results:
                # Track this read for combined processing
                combined_results_read_indices.append(read_index)
            else:
                # Process the read independently
                sub_context = test_context
                processor = get_processor(sub_context)
                measurement_groups, calculated_data = processor.process(sub_context)
                all_measurement_groups.append(measurement_groups)
                all_calculated_data.extend(calculated_data)

    # Process all reads that use the combined results section together (only once)
    if combined_results_read_indices:
        # Use the first read's context but with all read_data included
        first_read_index = combined_results_read_indices[0]
        first_read_data = context.read_data[first_read_index - 1]
        combined_context = reader.create_read_context(
            context, first_read_data, first_read_index
        )
        processor = get_processor(combined_context)
        measurement_groups, calculated_data = processor.process(combined_context)
        all_measurement_groups.append(measurement_groups)
        all_calculated_data.extend(calculated_data)

    # Merge measurement groups by well position
    merged_groups = merge_measurement_groups(all_measurement_groups)

    if not merged_groups:
        raise AllotropeConversionError(NO_MEASUREMENTS_ERROR)

    return merged_groups, all_calculated_data



from __future__ import annotations

from io import StringIO
import re
from typing import cast, ClassVar

import pandas as pd

from allotropy.exceptions import AllotropeConversionError, AllotropeParsingError
from allotropy.named_file_contents import NamedFileContents
from allotropy.parsers.constants import round_to_nearest_well_count
from allotropy.parsers.lines_reader import CsvReader, read_to_lines
from allotropy.parsers.luminex_xponent import constants
from allotropy.parsers.utils.pandas import read_csv
from allotropy.parsers.utils.values import assert_not_none, try_float_or_none


class LuminexXponentReader:
    SUPPORTED_EXTENSIONS = "csv"

    def __init__(self, named_file_contents: NamedFileContents) -> None:
        self.lines = read_to_lines(named_file_contents)

        if LuminexXponentReader._is_single_dataset(self.lines):
            (
                self.results_data,
                self.header_data,
                self.calibration_data,
                self.minimum_assay_bead_count_setting,
            ) = SingleDatasetParser.parse(self.lines)
        else:
            reader = CsvReader(self.lines)
            (
                self.results_data,
                self.header_data,
                self.calibration_data,
                self.minimum_assay_bead_count_setting,
            ) = MultipleDatasetParser.parse(reader)

    @staticmethod
    def _is_single_dataset(lines: list[str]) -> bool:
        first_line = lines[0] or ""
        return (
            "INSTRUMENT TYPE" in first_line
            and "WELL LOCATION" in first_line
            and "SAMPLE ID" in first_line
        )


class SingleDatasetParser:
    """
    Adapter to read single dataset CSV exports (all data in a single block)
    and yield the same data structures expected by xPONENT-based downstream code.
    """

    # Columns that are always present in the single-dataset CSV export and are not analyte columns
    FIXED_INPUT_COLUMNS: ClassVar[list[str]] = [
        "INSTRUMENT TYPE",
        "SERIAL NUMBER",
        "SOFTWARE VERSION",
        "PLATE NAME",
        "PLATE START",
        "WELL LOCATION",
        "SAMPLE ID",
    ]

    @staticmethod
    def parse_header(col: str) -> tuple[str, str] | None:
        """Parse a column header into (analyte_label, metric_token).

        The column names have an ID followed by the metric. The metric itself is
        composed of space-separated words that must all be included in
        constants.SINGLE_DATASET_RESULTS_METRIC_WORDS. We therefore split the
        header on spaces and find the earliest split index such that the suffix
        (metric) is entirely composed of known metric words (case-insensitive),
        assigning the prefix to the analyte label.
        """
        header = str(col).strip()
        parts = [p for p in header.split(" ") if p]
        if len(parts) < 2:
            return None

        # Compare words case-insensitively, without extra cleaning
        cleaned = [p.upper() for p in parts]

        # Find earliest index where the suffix is all known metric words
        split_index: int | None = None
        for i in range(1, len(parts)):
            suffix = cleaned[i:]
            if suffix and all(
                s in constants.SINGLE_DATASET_RESULTS_METRIC_WORDS for s in suffix
            ):
                split_index = i
                break

        if split_index is None:
            return None

        analyte_part = " ".join(parts[:split_index]).strip()
        metric_part = " ".join(parts[split_index:]).strip()
        if not analyte_part or not metric_part:
            return None
        return analyte_part, metric_part

    @staticmethod
    def section_name_from_metric(metric_token: str) -> str:
        """Map raw metric tokens to xPONENT-style section names."""
        token_upper = metric_token.upper().strip()
        if token_upper.endswith("AVERAGE MFI"):
            return "Avg Net MFI"
        # Title-case fallback for all other tokens
        return metric_token.title()

    @staticmethod
    def build_section(
        df: pd.DataFrame,
        analyte_labels: list[str],
        headers_map: dict[tuple[str, str], str],
        metric_token: str,
    ) -> pd.DataFrame | None:
        """Build a results section dataframe for the given metric.

        Returns None if no analyte column exists for this metric in the input.
        """

        if metric_token.upper().strip() == "UNITS":
            return SingleDatasetParser.build_units_section(
                df, analyte_labels, headers_map
            )

        at_least_one_present = False
        out = pd.DataFrame()
        out["Location"] = [f"1(1,{w})" for w in df["WELL LOCATION"].astype(str)]
        out["Sample"] = df["SAMPLE ID"].astype(str)
        for analyte in analyte_labels:
            src_col = headers_map.get((analyte, metric_token))
            if src_col:
                at_least_one_present = True
            out[analyte] = df[src_col] if src_col in df.columns else ""

        if not at_least_one_present:
            return None

        if metric_token == "COUNT":  # noqa: S105
            analyte_cols = analyte_labels
            # Convert each analyte column to numeric, coercing errors to NaN then filling with 0.0
            converted = cast(
                pd.DataFrame,
                out[analyte_cols].apply(pd.to_numeric, errors="coerce").fillna(0.0),
            )
            out["Total Events"] = converted.sum(axis=1)
        else:
            out["Total Events"] = ""
        return out.set_index("Location")

    @staticmethod
    def build_units_section(
        df: pd.DataFrame,
        analyte_labels: list[str],
        headers_map: dict[tuple[str, str], str],
    ) -> pd.DataFrame:
        """Build a units section dataframe for the given analyte labels."""
        units_list = []
        at_least_one_present = False
        for analyte in analyte_labels:
            src_col = headers_map.get((analyte, "UNITS"))
            units_list.append(df[src_col][0] if src_col in df.columns else "")
            if src_col:
                at_least_one_present = True

        if not at_least_one_present:
            return pd.DataFrame()

        columns = ["Analyte:", *analyte_labels]
        units: pd.DataFrame = pd.DataFrame(
            [
                columns,
                ["BeadID:", *units_list],
                ["Units:", *[""] * len(analyte_labels)],
            ],
            index=["Analyte:", "BeadID:", "Units:"],
            columns=columns,
        )

        return units

    @staticmethod
    def add_mandatory_columns(
        results: dict[str, pd.DataFrame],
        base_df: pd.DataFrame,
        analyte_labels: list[str],
    ) -> None:
        """Ensure required sections exist (Units, Dilution Factor)."""
        if "Dilution Factor" not in results:
            results["Dilution Factor"] = pd.DataFrame()
            results["Dilution Factor"]["Location"] = [
                f"1(1,{w})" for w in base_df["WELL LOCATION"].astype(str)
            ]
            results["Dilution Factor"]["Sample"] = base_df["SAMPLE ID"].astype(str)
            results["Dilution Factor"]["Dilution Factor"] = ""
            results["Dilution Factor"] = results["Dilution Factor"].set_index(
                "Location"
            )
        if "Units" not in results:
            units_df = pd.DataFrame(
                [
                    ["Analyte:", *analyte_labels],
                    ["BeadID:"],
                    ["Units:"],
                ],
                index=["Analyte:", "BeadID:", "Units:"],
                columns=["Analyte:", *analyte_labels],
            )
            results["Units"] = units_df

    @staticmethod
    def _get_well_count(df: pd.DataFrame) -> int:
        # Count unique values in WELL LOCATION matching Excel-like cell ids: Letter(s) + Number(s)
        # Example matches: A1, B12, AA3
        pattern = r"^[A-Za-z]+\d+$"
        values = df["WELL LOCATION"].astype(str).str.strip()
        matching = values[values.str.match(pattern)]
        unique_count = int(matching.str.upper().nunique())
        nearest = round_to_nearest_well_count(
            unique_count, constants.POSSIBLE_WELL_COUNTS_LUMINEX
        )
        return int(nearest or 0)

    @classmethod
    def parse(
        cls, lines: list[str]
    ) -> tuple[dict[str, pd.DataFrame], pd.DataFrame, pd.DataFrame, float | None]:
        """Parse single-dataset CSV and return header, calibration, min beads and results tables."""
        df = read_csv(StringIO("\n".join(lines)), header=0)

        analyte_labels: list[str] = []
        seen_labels: set[str] = set()
        headers_map: dict[tuple[str, str], str] = {}
        metric_tokens_ordered: list[str] = []
        seen_metric_tokens: set[str] = set()
        for col in df.columns:
            if col in cls.FIXED_INPUT_COLUMNS:
                continue
            parsed = cls.parse_header(str(col))
            if not parsed:
                continue
            analyte_label, metric = parsed
            headers_map[(analyte_label, metric)] = str(col)
            if (
                metric.upper().strip().endswith("COUNT")
                and analyte_label not in seen_labels
            ):
                analyte_labels.append(analyte_label)
                seen_labels.add(analyte_label)
            if metric not in seen_metric_tokens:
                metric_tokens_ordered.append(metric)
                seen_metric_tokens.add(metric)

        results: dict[str, pd.DataFrame] = {}

        for token in metric_tokens_ordered:
            section_name = cls.section_name_from_metric(token)
            section_df = cls.build_section(df, analyte_labels, headers_map, token)
            if section_df is not None:
                results[section_name] = section_df

        cls.add_mandatory_columns(results, df, analyte_labels)

        # Safely extract optional fields from the input DataFrame without mypy complaints
        _serial_series = (
            df["SERIAL NUMBER"] if "SERIAL NUMBER" in df.columns else pd.Series([""])
        )
        _serial_value = str(_serial_series.iloc[0]) if len(_serial_series) > 0 else ""
        _plate_series = (
            df["PLATE START"] if "PLATE START" in df.columns else pd.Series([""])
        )
        _plate_value = str(_plate_series.iloc[0]) if len(_plate_series) > 0 else ""

        header = (
            pd.DataFrame(
                {
                    0: [
                        "Program",
                        "Build",
                        "Date",
                        "SN",
                        "ProtocolPlate",
                        "ComputerName",
                        "BatchStartTime",
                    ],
                    1: [
                        "xPONENT",
                        "Unknown",
                        "",
                        _serial_value,
                        "Name",
                        "",
                        _plate_value,
                    ],
                    2: ["", "", "", "", "", "", ""],
                    3: ["", "", "", "", "", "", ""],
                    4: ["", "", "", "", cls._get_well_count(df), "", ""],
                    5: ["", "", "", "", "", "", ""],
                }
            )
            .set_index(0)
            .T
        )

        calibration = pd.DataFrame()
        min_beads: float | None = None

        return results, header, calibration, min_beads


class MultipleDatasetParser:
    @classmethod
    def parse(
        cls, reader: CsvReader
    ) -> tuple[dict[str, pd.DataFrame], pd.DataFrame, pd.DataFrame, float | None]:

        header_data = cls._get_header_data(reader)
        calibration_data = cls._get_calibration_data(reader)
        minimum_assay_bead_count_setting = cls._get_minimum_assay_bead_count_setting(
            reader
        )
        results_data = cls._get_results(reader)
        return (
            results_data,
            header_data,
            calibration_data,
            minimum_assay_bead_count_setting,
        )

    @classmethod
    def _get_header_data(cls, reader: CsvReader) -> pd.DataFrame:
        header_lines = list(reader.pop_until(constants.CALIBRATION_BLOCK_HEADER))
        # Header has a weird structure where there rows have varying number of columns, so we need
        # special handling before passing to read_csv.
        n_columns = max(len(line.split(",")) for line in header_lines)
        if n_columns < constants.EXPECTED_HEADER_COLUMNS:
            msg = "Unable to parse header. Not enough data."
            raise AllotropeConversionError(msg)

        return (
            read_csv(
                StringIO("\n".join(header_lines)),
                header=None,
                index_col=0,
                names=range(n_columns),
            )
            .dropna(how="all")
            .T
        )

    @classmethod
    def _get_calibration_data(cls, reader: CsvReader) -> pd.DataFrame:
        reader.drop_until_inclusive(constants.CALIBRATION_BLOCK_HEADER)
        return assert_not_none(
            reader.pop_csv_block_as_df(constants.LUMINEX_EMPTY_PATTERN)
        )

    @classmethod
    def _get_minimum_assay_bead_count_setting(cls, reader: CsvReader) -> float | None:
        reader.drop_until(match_pat='^"?Samples"?,')
        samples_info = assert_not_none(reader.pop(), msg="Unable to find Samples info.")
        try:
            fields = samples_info.replace('"', "").split(",")
            min_bead_count_setting = fields[3].strip()
            # If the min bead count is left empty, the default value is 100, according to software manual.
            if not min_bead_count_setting:
                return 100
        except IndexError as e:
            msg = f"Unable to find minimum bead count setting in Samples info: {samples_info}."
            raise AllotropeConversionError(msg) from e

        return try_float_or_none(min_bead_count_setting)

    @staticmethod
    def _get_results(reader: CsvReader) -> dict[str, pd.DataFrame]:
        reader.drop_until_inclusive("Results")
        reader.drop_empty(constants.LUMINEX_EMPTY_PATTERN)
        results: dict[str, pd.DataFrame] = {}
        while reader.current_line_exists() and "-- CRC --" not in (reader.get() or ""):
            result_title_line = assert_not_none(reader.pop())
            match: re.Match[str] | None
            if not (
                match := re.match(constants.TABLE_HEADER_PATTERN, result_title_line)
            ):
                msg = f"Invalid header block start line: {result_title_line}"
                raise AllotropeParsingError(msg)
            result_title = match.groups()[0]
            table_data = assert_not_none(
                reader.pop_csv_block_as_df(
                    empty_pat=constants.LUMINEX_EMPTY_PATTERN,
                    header=0,
                    index_col=0,
                    na_values=["NaN", "nan", "-NaN", "-nan", "None", "null", "NULL"],
                    keep_default_na=False,  # Prevent pandas from interpreting empty strings as NaN.
                )
            )
            # drop empty columns at the end of the table
            table_data = table_data.loc[:, ~table_data.columns.str.contains("^Unnamed")]
            results[result_title] = table_data
            reader.drop_empty(constants.LUMINEX_EMPTY_PATTERN)

        return results



# generated by datamodel-codegen:
#   filename:  mass-spectrometry.schema.json
#   timestamp: 2025-11-05T18:53:18+00:00

from __future__ import annotations

from dataclasses import dataclass

from allotropy.allotrope.models.shared.definitions.custom import (
    TQuantityValueDalton,
    TQuantityValueHertz,
    TQuantityValueMassPerCharge,
    TQuantityValueMilliliterPerMinute,
    TQuantityValuePercent,
    TQuantityValueSecondTime,
)
from allotropy.allotrope.models.shared.definitions.definitions import (
    TClass,
    TQuantityValue,
    TStringValue,
)

TNamed = str


@dataclass(kw_only=True)
class DeviceDocumentItem:
    device_type: TClass
    model_number: TStringValue
    product_manufacturer: TStringValue


@dataclass(kw_only=True)
class DeviceSystemDocument:
    asset_management_identifier: TStringValue
    device_document: list[DeviceDocumentItem] | None = None


@dataclass(kw_only=True)
class DataProcessingDocumentItem:
    data_processing_type: TStringValue | None = None


@dataclass(kw_only=True)
class DataProcessingAggregateDocument:
    data_processing_document: list[DataProcessingDocumentItem] | None = None


@dataclass(kw_only=True)
class DetectorControlDocumentItem:
    detection_type: TStringValue
    detection_duration_setting: TQuantityValueSecondTime | None = None
    detector_relative_offset_setting: TQuantityValueSecondTime | None = None
    detector_sampling_rate_setting: TQuantityValueHertz | None = None
    m_z_maximum_setting: TQuantityValueMassPerCharge | None = None
    m_z_minimum_setting: TQuantityValueMassPerCharge | None = None
    polarity_setting: TNamed | None = None
    field_index: int | None = None


@dataclass(kw_only=True)
class DetectorControlAggregateDocument:
    detector_control_document: list[DetectorControlDocumentItem] | None = None


@dataclass(kw_only=True)
class MeasurementDocument:
    measurement_mode_setting: TStringValue
    detector_control_aggregate_document: DetectorControlAggregateDocument | None = None


@dataclass(kw_only=True)
class ProcessedDataDocumentItem:
    data_format_specification_type: TStringValue


@dataclass(kw_only=True)
class ProcessedDataAggregateDocument:
    processed_data_document: list[ProcessedDataDocumentItem] | None = None


@dataclass(kw_only=True)
class SampleDocument:
    sample_identifier: TStringValue


@dataclass(kw_only=True)
class SampleIntroductionDocument:
    sample_introduction_medium: TStringValue
    sample_introduction_mode_setting: TStringValue
    flow_rate_setting: TQuantityValueMilliliterPerMinute | None = None
    laser_firing_frequency_setting: TQuantityValueHertz | None = None
    sample_introduction_description: TStringValue | None = None


@dataclass(kw_only=True)
class MassSpectrometryDocumentItem:
    analyst: TStringValue
    submitter: TStringValue
    data_processing_aggregate_document: DataProcessingAggregateDocument | None = None
    measurement_document: MeasurementDocument | None = None
    processed_data_aggregate_document: ProcessedDataAggregateDocument | None = None
    sample_document: SampleDocument | None = None
    sample_introduction_document: SampleIntroductionDocument | None = None


@dataclass(kw_only=True)
class MassSpectrometryAggregateDocument:
    device_system_document: DeviceSystemDocument | None = None
    mass_spectrometry_document: list[MassSpectrometryDocumentItem] | None = None


@dataclass(kw_only=True)
class PeakItem:
    identifier: TStringValue | None = None
    m_z: TQuantityValueMassPerCharge | None = None
    mass: TQuantityValueDalton | None = None
    peak_area: TQuantityValue | None = None
    peak_height: TQuantityValue | None = None
    peak_width: TQuantityValue | None = None
    relative_peak_area: TQuantityValuePercent | None = None
    relative_peak_height: TQuantityValuePercent | None = None
    written_name: TStringValue | None = None
    field_index: int | None = None


@dataclass(kw_only=True)
class PeakList:
    peak: list[PeakItem] | None = None


@dataclass(kw_only=True)
class Model:
    manifest: str = "http://purl.allotrope.org/manifests/mass-spectrometry/REC/2025/06/mass-spectrometry.manifest"
    data_processing_description: TStringValue | None = None
    mass_spectrometry_aggregate_document: MassSpectrometryAggregateDocument | None = (
        None
    )
    peak_list: PeakList | None = None



"""Constants file for Agilent OpenLab CDS Adapter"""

from allotropy.allotrope.models.shared.components.plate_reader import SampleRoleType

DISPLAY_NAME = "Agilent OpenLab CDS"
PRODUCT_MANUFACTURER = "Agilent"
SAMPLE_ROLE_TYPE = {
    "Blank": SampleRoleType.blank_role.value,
    "Sample": SampleRoleType.sample_role.value,
}



from enum import Enum

import pandas as pd

from allotropy.allotrope.schema_mappers.adm.plate_reader.rec._2024._06.plate_reader import (
    Measurement,
)
from allotropy.calcdocs.config import (
    CalcDocsConfig,
    CalculatedDataConfig,
    MeasurementConfig,
)
from allotropy.calcdocs.msd_workbench.extractor import MsdWorkbenchExtractor
from allotropy.calcdocs.msd_workbench.views import (
    AssayIdentifierView,
    LocationIdentifierView,
    SampleIdentifierView,
    WellPlateIdentifierView,
)
from allotropy.parsers.msd_workbench.calculdated_data_structure import (
    CalculatedDataMeasurementStructure,
)
from allotropy.parsers.utils.calculated_data_documents.definition import (
    CalculatedDocument,
)
from allotropy.parsers.utils.pandas import SeriesData


class AggregatingProperty(Enum):
    ASSAY_IDENTIFIER = "assay identifier"
    SAMPLE_IDENTIFIER = "sample_identifier"
    WELL_PLATE_IDENTIFIER = "well_plate_identifier"
    LOCATION_IDENTIFIER = "location_identifier"
    WELL_PLATE_ID = "well_plate_identifier"


class CalculatedDataColumns(Enum):
    ADJUSTED_SIGNAL = "Adjusted Signal"
    MEAN = "Mean"
    ADJ_SIG_MEAN = "Adj. Sig. Mean"
    FIT_STATISTIC_RSQUARED = "Fit Statistic: RSquared"
    CV = "CV"
    PERCENT_RECOVERY = "% Recovery"
    PERCENT_RECOVERY_MEAN = "% Recovery Mean"
    CALC_CONCENTRATION = "Calc. Concentration"
    CALC_CONC_MEAN = "Calc. Conc. Mean"
    CALC_CONC_CV = "Calc. Conc. CV"


def create_calculated_data_groups(
    data: pd.DataFrame, measurements: list[Measurement]
) -> list[CalculatedDocument]:
    calc_data_measurements: list[CalculatedDataMeasurementStructure] = []
    for measurement in measurements:
        for _, row in data.iterrows():
            row_series = SeriesData(row)
            if measurement.well_location_identifier != row_series.get(
                str, "Well"
            ) or measurement.location_identifier != row_series.get(str, "Spot"):
                row_series.get_unread()
                continue
            calc_data_measurements.append(
                CalculatedDataMeasurementStructure(
                    measurement=measurement,
                    adjusted_signal=row_series.get(
                        float, CalculatedDataColumns.ADJUSTED_SIGNAL.value
                    ),
                    mean=row_series.get(float, CalculatedDataColumns.MEAN.value),
                    adj_sig_mean=row_series.get(
                        float, CalculatedDataColumns.ADJ_SIG_MEAN.value
                    ),
                    fit_statistic_rsquared=row_series.get(
                        float, CalculatedDataColumns.FIT_STATISTIC_RSQUARED.value
                    ),
                    cv=row_series.get(float, CalculatedDataColumns.CV.value),
                    percent_recovery=row_series.get(
                        float, CalculatedDataColumns.PERCENT_RECOVERY.value
                    ),
                    percent_recovery_mean=row_series.get(
                        float, CalculatedDataColumns.PERCENT_RECOVERY_MEAN.value
                    ),
                    calc_concentration=row_series.get(
                        float, CalculatedDataColumns.CALC_CONCENTRATION.value
                    ),
                    calc_conc_mean=row_series.get(
                        float, CalculatedDataColumns.CALC_CONC_MEAN.value
                    ),
                    calc_conc_cv=row_series.get(
                        float, CalculatedDataColumns.CALC_CONC_CV.value
                    ),
                )
            )
            # we do not need additional data for the calculated data documents
            row_series.get_unread()
    elements = MsdWorkbenchExtractor.get_elements(calc_data_measurements)
    sample_assay_view = SampleIdentifierView(sub_view=AssayIdentifierView()).apply(
        elements
    )
    assay_view = AssayIdentifierView().apply(elements)
    sample_well_plate_location_id_view = AssayIdentifierView(
        sub_view=WellPlateIdentifierView(
            sub_view=LocationIdentifierView(SampleIdentifierView())
        )
    ).apply(elements)
    measurement_conf = MeasurementConfig(
        name="luminescence",
        value="luminescence",
    )
    adj_signal_config = CalculatedDataConfig(
        name=CalculatedDataColumns.ADJUSTED_SIGNAL.value,
        value="adjusted_signal",
        view_data=sample_assay_view,
        source_configs=(measurement_conf,),
    )
    mean_config = CalculatedDataConfig(
        name=CalculatedDataColumns.MEAN.value,
        value="mean",
        view_data=assay_view,
        source_configs=(measurement_conf,),
    )
    adj_signal_mean_config = CalculatedDataConfig(
        name=CalculatedDataColumns.ADJ_SIG_MEAN.value,
        value="adj_sig_mean",
        view_data=assay_view,
        source_configs=(mean_config,),
    )
    rsquared_config = CalculatedDataConfig(
        name="R-Squared",
        value="fit_statistic_rsquared",
        view_data=assay_view,
        source_configs=(measurement_conf,),
    )
    cv_config = CalculatedDataConfig(
        name=CalculatedDataColumns.CV.value,
        value="cv",
        view_data=assay_view,
        source_configs=(measurement_conf,),
    )
    percent_recovery_config = CalculatedDataConfig(
        name=CalculatedDataColumns.PERCENT_RECOVERY.value,
        value="percent_recovery",
        view_data=sample_well_plate_location_id_view,
        source_configs=(measurement_conf,),
    )
    percent_recovery_mean_config = CalculatedDataConfig(
        name=CalculatedDataColumns.PERCENT_RECOVERY_MEAN.value,
        value="percent_recovery_mean",
        view_data=assay_view,
        source_configs=(percent_recovery_config,),
    )
    calc_concentration_config = CalculatedDataConfig(
        name=CalculatedDataColumns.CALC_CONCENTRATION.value,
        value="calc_concentration",
        view_data=sample_well_plate_location_id_view,
        source_configs=(measurement_conf,),
    )
    calc_conc_mean_config = CalculatedDataConfig(
        name="Calc. Concentration Mean",
        value="calc_conc_mean",
        view_data=assay_view,
        source_configs=(calc_concentration_config,),
    )
    calc_conc_cv_config = CalculatedDataConfig(
        name=CalculatedDataColumns.CALC_CONC_CV.value,
        value="calc_conc_cv",
        view_data=assay_view,
        source_configs=(calc_concentration_config,),
    )
    configs = CalcDocsConfig(
        [
            mean_config,
            adj_signal_config,
            adj_signal_mean_config,
            rsquared_config,
            cv_config,
            percent_recovery_config,
            percent_recovery_mean_config,
            calc_concentration_config,
            calc_conc_mean_config,
            calc_conc_cv_config,
        ]
    )
    calc_docs = [
        calc_doc
        for parent_calc_doc in configs.construct()
        for calc_doc in parent_calc_doc.iter_struct()
    ]
    return calc_docs



from __future__ import annotations

from dataclasses import dataclass
from typing import Any

from allotropy.allotrope.converter import add_custom_information_document
from allotropy.allotrope.models.adm.flow_cytometry.benchling._2025._03.flow_cytometry import (
    CompensationMatrixAggregateDocument,
    CompensationMatrixDocumentItem,
    DataProcessingDocument,
    DataRegionAggregateDocument,
    DataRegionDocumentItem,
    DataSystemDocument,
    DeviceControlAggregateDocument,
    DeviceControlDocumentItem,
    DeviceSystemDocument,
    FlowCytometryAggregateDocument,
    FlowCytometryDocumentItem,
    MatrixAggregateDocument,
    MatrixDocumentItem,
    MeasurementAggregateDocument,
    MeasurementDocumentItem,
    Model,
    PopulationAggregateDocumentItem,
    PopulationDocumentItem,
    ProcessedDataAggregateDocument,
    ProcessedDataDocumentItem,
    SampleDocument,
    StatisticDimensionAggregateDocument,
    StatisticDimensionDocumentItem,
    StatisticsAggregateDocument,
    StatisticsDocumentItem,
    VertexAggregateDocument,
    VertexDocumentItem,
)
from allotropy.allotrope.models.shared.definitions.custom import (
    TQuantityValueCounts,
    TQuantityValueUnitless,
)
from allotropy.allotrope.models.shared.definitions.definitions import (
    TQuantityValue,
    TStatisticDatumRole,
)
from allotropy.allotrope.schema_mappers.schema_mapper import SchemaMapper
from allotropy.constants import ASM_CONVERTER_VERSION
from allotropy.parsers.utils.values import quantity_or_none, quantity_or_none_from_unit


@dataclass(frozen=True)
class Vertex:
    x_coordinate: float
    y_coordinate: float
    x_unit: str
    y_unit: str
    vertex_role: str | None = None
    custom_info: dict[str, Any] | None = None


@dataclass(frozen=True)
class DataRegion:
    region_data_identifier: str | None
    region_data_type: str | None
    parent_data_region_identifier: str | None = None
    x_coordinate_dimension_identifier: str | None = None
    y_coordinate_dimension_identifier: str | None = None
    vertices: list[Vertex] | None = None
    custom_info: dict[str, Any] | None = None


@dataclass(frozen=True)
class CompensationMatrix:
    dimension_identifier: str | None
    compensation_value: float | None
    custom_info: dict[str, Any] | None = None


@dataclass(frozen=True)
class CompensationMatrixGroup:
    dimension_identifier: str | None
    compensation_matrices: list[CompensationMatrix] | None = None
    custom_info: dict[str, Any] | None = None


@dataclass(frozen=True)
class StatisticDimension:
    value: float
    unit: str
    has_statistic_datum_role: str | None = None
    dimension_identifier: str | None = None
    custom_info: dict[str, Any] | None = None


@dataclass(frozen=True)
class Statistic:
    statistic_dimension: list[StatisticDimension]
    statistical_feature: str
    custom_info: dict[str, Any] | None = None


@dataclass(frozen=True)
class Population:
    population_identifier: str
    written_name: str | None = None
    data_region_identifier: str | None = None
    count: int | None = None
    parent_population_identifier: str | None = None
    sub_populations: list[Population] | None = None
    statistics: list[Statistic] | None = None
    custom_info: dict[str, Any] | None = None


@dataclass(frozen=True)
class Measurement:
    measurement_identifier: str
    sample_identifier: str
    device_type: str

    location_identifier: str | None = None
    well_plate_identifier: str | None = None
    batch_identifier: str | None = None
    written_name: str | None = None
    method_version: str | None = None
    data_processing_time: str | None = None
    processed_data_identifier: str | None = None
    populations: list[Population] | None = None
    data_regions: list[DataRegion] | None = None

    sample_custom_info: dict[str, Any] | None = None
    processed_data_custom_info: dict[str, Any] | None = None
    data_processing_custom_info: dict[str, Any] | None = None
    device_control_custom_info: dict[str, Any] | None = None
    custom_info: dict[str, Any] | None = None


@dataclass(frozen=True)
class MeasurementGroup:
    measurements: list[Measurement]
    compensation_matrix_groups: list[CompensationMatrixGroup] | None = None
    analyst: str | None = None
    measurement_time: str | None = None
    experimental_data_identifier: str | None = None
    experiment_identifier: str | None = None
    measurement_aggregate_custom_info: dict[str, Any] | None = None
    custom_info: dict[str, Any] | None = None


@dataclass
class Metadata:
    asm_file_identifier: str
    data_system_instance_identifier: str
    model_number: str | None = None
    software_name: str | None = None
    file_name: str | None = None
    unc_path: str | None = None
    equipment_serial_number: str | None = None
    software_version: str | None = None
    device_identifier: str | None = None

    custom_info: dict[str, Any] | None = None


@dataclass
class Data:
    metadata: Metadata
    measurement_groups: list[MeasurementGroup]


class Mapper(SchemaMapper[Data, Model]):
    MANIFEST = "http://purl.allotrope.org/manifests/flow-cytometry/BENCHLING/2025/03/flow-cytometry.manifest"

    def map_model(self, data: Data) -> Model:
        return Model(
            manifest=self.MANIFEST,
            flow_cytometry_aggregate_document=add_custom_information_document(
                FlowCytometryAggregateDocument(
                    device_system_document=DeviceSystemDocument(
                        device_identifier=data.metadata.device_identifier,
                        model_number=data.metadata.model_number,
                        equipment_serial_number=data.metadata.equipment_serial_number,
                    ),
                    data_system_document=DataSystemDocument(
                        data_system_instance_identifier=data.metadata.data_system_instance_identifier,
                        file_name=data.metadata.file_name,
                        UNC_path=data.metadata.unc_path,
                        software_name=data.metadata.software_name,
                        software_version=data.metadata.software_version,
                        ASM_converter_name=self.converter_name,
                        ASM_converter_version=ASM_CONVERTER_VERSION,
                        ASM_file_identifier=data.metadata.asm_file_identifier,
                    ),
                    flow_cytometry_document=[
                        self._get_technique_document(measurement_group)
                        for measurement_group in data.measurement_groups
                    ],
                ),
                data.metadata.custom_info,
            ),
        )

    def _get_technique_document(
        self, measurement_group: MeasurementGroup
    ) -> FlowCytometryDocumentItem:
        return add_custom_information_document(
            FlowCytometryDocumentItem(
                measurement_aggregate_document=add_custom_information_document(
                    MeasurementAggregateDocument(
                        measurement_document=[
                            self._get_measurement_document(measurement)
                            for measurement in measurement_group.measurements
                        ],
                        analyst=measurement_group.analyst,
                        measurement_time=self.get_date_time(
                            measurement_group.measurement_time
                        )
                        if measurement_group.measurement_time
                        else None,
                        experimental_data_identifier=measurement_group.experimental_data_identifier,
                        experiment_identifier=measurement_group.experiment_identifier,
                    ),
                    measurement_group.measurement_aggregate_custom_info,
                ),
                compensation_matrix_aggregate_document=CompensationMatrixAggregateDocument(
                    compensation_matrix_document=[
                        self._get_compensation_matrix_document(compensation_matrix)
                        for compensation_matrix in measurement_group.compensation_matrix_groups
                    ]
                    if measurement_group.compensation_matrix_groups
                    else None,
                ),
            ),
            measurement_group.custom_info,
        )

    def _get_measurement_document(
        self, measurement: Measurement
    ) -> MeasurementDocumentItem:
        return add_custom_information_document(
            MeasurementDocumentItem(
                measurement_identifier=measurement.measurement_identifier,
                sample_document=self._get_sample_document(measurement),
                device_control_aggregate_document=DeviceControlAggregateDocument(
                    device_control_document=[
                        add_custom_information_document(
                            DeviceControlDocumentItem(
                                device_type=measurement.device_type,
                            ),
                            measurement.device_control_custom_info,
                        )
                    ],
                ),
                processed_data_aggregate_document=ProcessedDataAggregateDocument(
                    processed_data_document=[
                        add_custom_information_document(
                            ProcessedDataDocumentItem(
                                data_processing_document=add_custom_information_document(
                                    DataProcessingDocument(
                                        method_version=measurement.method_version,
                                        data_processing_time=self.get_date_time(
                                            measurement.data_processing_time
                                        )
                                        if measurement.data_processing_time
                                        else None,
                                    ),
                                    measurement.data_processing_custom_info,
                                ),
                                processed_data_identifier=measurement.processed_data_identifier,
                                population_aggregate_document=[
                                    PopulationAggregateDocumentItem(
                                        population_document=[
                                            self._get_population_document(population)
                                            for population in measurement.populations
                                        ]
                                        if measurement.populations
                                        else None
                                    )
                                ],
                                data_region_aggregate_document=DataRegionAggregateDocument(
                                    data_region_document=[
                                        self._get_data_region_document(data_region)
                                        for data_region in measurement.data_regions
                                    ]
                                    if measurement.data_regions
                                    else None
                                ),
                            ),
                            measurement.processed_data_custom_info,
                        )
                    ]
                ),
            ),
            measurement.custom_info,
        )

    def _get_data_region_document(
        self, data_region: DataRegion
    ) -> DataRegionDocumentItem:
        return add_custom_information_document(
            DataRegionDocumentItem(
                data_region_identifier=data_region.region_data_identifier,
                region_type=data_region.region_data_type,
                parent_data_region_identifier=data_region.parent_data_region_identifier,
                x_coordinate_dimension_identifier=data_region.x_coordinate_dimension_identifier,
                y_coordinate_dimension_identifier=data_region.y_coordinate_dimension_identifier,
                vertex_aggregate_document=self._get_vertex_aggregate_document(
                    data_region.vertices
                ),
            ),
            data_region.custom_info,
        )

    def _get_vertex_aggregate_document(
        self, vertices: list[Vertex] | None
    ) -> VertexAggregateDocument | None:
        if not vertices:
            return None
        return VertexAggregateDocument(
            vertex_document=[
                add_custom_information_document(
                    VertexDocumentItem(
                        x_coordinate=quantity_or_none_from_unit(vertex.x_unit, vertex.x_coordinate),  # type: ignore[arg-type]
                        y_coordinate=quantity_or_none_from_unit(vertex.y_unit, vertex.y_coordinate),  # type: ignore[arg-type]
                        vertex_role=vertex.vertex_role,
                    ),
                    vertex.custom_info,
                )
                for vertex in vertices
            ]
        )

    def _get_population_document(
        self, population: Population
    ) -> PopulationDocumentItem:
        return add_custom_information_document(
            PopulationDocumentItem(
                population_identifier=population.population_identifier,
                written_name=population.written_name,
                data_region_identifier=population.data_region_identifier,
                count=quantity_or_none(TQuantityValueCounts, value=population.count),
                parent_population_identifier=population.parent_population_identifier,
                population_aggregate_document=[
                    PopulationAggregateDocumentItem(
                        population_document=[
                            self._get_population_document(sub_population)
                            for sub_population in population.sub_populations
                        ]
                    )
                ]
                if population.sub_populations
                else None,
                statistics_aggregate_document=self._get_statistics_aggregate_document(
                    population.statistics
                ),
            ),
            population.custom_info,
        )

    def _get_compensation_matrix_document(
        self, compensation_matrix_group: CompensationMatrixGroup
    ) -> CompensationMatrixDocumentItem:
        return add_custom_information_document(
            CompensationMatrixDocumentItem(
                dimension_identifier=compensation_matrix_group.dimension_identifier,
                matrix_aggregate_document=MatrixAggregateDocument(
                    matrix_document=[
                        add_custom_information_document(
                            MatrixDocumentItem(
                                dimension_identifier=compensation.dimension_identifier,
                                compensation_value=quantity_or_none(
                                    TQuantityValueUnitless,
                                    compensation.compensation_value,
                                ),
                            ),
                            compensation.custom_info,
                        )
                        for compensation in compensation_matrix_group.compensation_matrices
                    ]
                    if compensation_matrix_group.compensation_matrices
                    else None,
                ),
            ),
            compensation_matrix_group.custom_info,
        )

    def _get_statistics_aggregate_document(
        self, statistics: list[Statistic] | None
    ) -> StatisticsAggregateDocument | None:
        if not statistics:
            return None
        return StatisticsAggregateDocument(
            statistics_document=[
                add_custom_information_document(
                    StatisticsDocumentItem(
                        statistical_feature=statistic.statistical_feature,
                        statistic_dimension_aggregate_document=StatisticDimensionAggregateDocument(
                            statistic_dimension_document=[
                                add_custom_information_document(
                                    StatisticDimensionDocumentItem(
                                        dimension_identifier=statistic_dimension.dimension_identifier,
                                        statistical_value=TQuantityValue(
                                            value=statistic_dimension.value,
                                            unit=statistic_dimension.unit,
                                            has_statistic_datum_role=TStatisticDatumRole(
                                                statistic_dimension.has_statistic_datum_role
                                            )
                                            if statistic_dimension.has_statistic_datum_role
                                            in {
                                                role.value
                                                for role in TStatisticDatumRole
                                            }
                                            else None,
                                        ),
                                    ),
                                    statistic_dimension.custom_info,
                                )
                                for statistic_dimension in statistic.statistic_dimension
                            ]
                        ),
                    ),
                    statistic.custom_info,
                )
                for statistic in statistics
            ]
        )

    def _get_sample_document(self, measurement: Measurement) -> SampleDocument:
        return add_custom_information_document(
            SampleDocument(
                sample_identifier=measurement.sample_identifier,
                location_identifier=measurement.location_identifier,
                well_plate_identifier=measurement.well_plate_identifier,
                written_name=measurement.written_name,
                batch_identifier=measurement.batch_identifier,
            ),
            measurement.sample_custom_info,
        )



from collections.abc import Mapping
import copy
import json
from pathlib import Path
from typing import Any

import jsonschema

from allotropy.allotrope.schema_parser.path_util import (
    get_full_schema_path,
    get_schema_path_from_asm,
    get_schema_path_from_manifest,
    SHARED_SCHEMAS_DEFINITIONS_PATH,
)
from allotropy.exceptions import AllotropeSerializationError, AllotropeValidationError

DEFAULT_ENCODING = "UTF-8"

# Override format checker to remove "uri-reference" check, which ASM schemas fail against.
FORMAT_CHECKER = copy.deepcopy(
    jsonschema.validators.Draft202012Validator.FORMAT_CHECKER
)
FORMAT_CHECKER.checkers.pop("uri-reference", None)


def get_shared_definitions() -> dict[str, Any]:
    with open(
        Path(SHARED_SCHEMAS_DEFINITIONS_PATH, "definitions.json"),
        encoding=DEFAULT_ENCODING,
    ) as f:
        return json.load(f)  # type: ignore[no-any-return]


def get_shared_unit_definitions() -> dict[str, Any]:
    with open(
        Path(SHARED_SCHEMAS_DEFINITIONS_PATH, "units.json"), encoding=DEFAULT_ENCODING
    ) as f:
        return json.load(f)  # type: ignore[no-any-return]


def add_definitions(schema: dict[str, Any]) -> dict[str, Any]:
    for file, section in [
        ("definitions", "defs"),
        ("units", "defs"),
        ("custom", "custom"),
    ]:
        existing = schema.get(f"${section}", {})
        with open(
            Path(SHARED_SCHEMAS_DEFINITIONS_PATH, f"{file}.json"),
            encoding=DEFAULT_ENCODING,
        ) as f:
            additional = json.load(f)
        additional.update(existing)
        schema[f"${section}"] = additional
    return schema


def get_schema(schema_path: Path) -> dict[str, Any]:
    with open(get_full_schema_path(schema_path), encoding=DEFAULT_ENCODING) as f:
        return add_definitions(json.load(f))


def get_schema_from_manifest(manifest: str) -> dict[str, Any]:
    return get_schema(get_schema_path_from_manifest(manifest))


def get_schema_from_asm(asm_dict: Mapping[str, Any]) -> dict[str, Any]:
    return get_schema(get_schema_path_from_asm(asm_dict))


def get_schema_from_model(model: Any) -> dict[str, Any]:
    manifest = getattr(model, "manifest", getattr(model, "field_asm_manifest", None))
    if not manifest:
        msg = f"No 'manifest' or 'field_asm_manifest' found in model: {type(model)}"
        raise ValueError(msg)
    return get_schema_from_manifest(manifest)


def validate_asm_schema(asm_dict: dict[str, Any]) -> None:
    try:
        allotrope_schema = get_schema_from_asm(asm_dict)
    except Exception as e:
        msg = f"Failed to retrieve schema for model: {e}"
        raise AllotropeSerializationError(msg) from e

    try:
        jsonschema.validators.Draft202012Validator.check_schema(
            allotrope_schema, format_checker=FORMAT_CHECKER
        )
        validator = jsonschema.validators.Draft202012Validator(allotrope_schema)
        validator.validate(asm_dict)
    except Exception as e:
        msg = f"Failed to validate allotrope model against schema: {e}"
        raise AllotropeValidationError(msg) from e
